# ActivitySim
# See full license in LICENSE.txt.

import logging
from builtins import object, range

import numpy as np
import pandas as pd

from activitysim.core.util import quick_loc_series

logger = logging.getLogger(__name__)

NOT_IN_SKIM_ZONE_ID = -1
NOT_IN_SKIM_NAN = np.nan

ROW_MAJOR_LAYOUT = True


class OffsetMapper(object):
    """
    Utility to map skim zone ids to ordinal offsets (e.g. numpy array indices)

    Can map either by a fixed offset (e.g. -1 to map 1-based to 0-based)
    or by an explicit mapping of zone id to offset (slower but more flexible)

    Internally, there are two representations:

    offset_int:
        int offset which when added to zone_id yields skim array index (e.g. -1 to map 1-based zones to 0-based index)
    offset_series:
        pandas series with zone_id index and skim array offset values. Ordinarily, index is just range(0, omx_size)
        if series has duplicate offset values, this can map multiple zone_ids to a single skim array index
        (e.g. can map maz zone_ids to corresponding taz skim offset)
    """

    def __init__(self, offset_int=None, offset_list=None, offset_series=None):

        self.offset_int = self.offset_series = None

        assert (offset_int is not None) + (offset_list is not None) + (
            offset_series is not None
        ) <= 1

        if offset_int is not None:
            self.set_offset_int(offset_int)
        elif offset_list is not None:
            self.set_offset_list(offset_list)
        elif offset_series is not None:
            self.set_offset_series(offset_series)

    def print_offset(self, message=""):
        assert (self.offset_int is not None) or (self.offset_series is not None)

        if self.offset_int is not None:
            print(f"{message} offset_int: {self.offset_int}")
        elif self.offset_series is not None:
            print(f"{message} offset_series:\n {self.offset_series}")
        else:
            print(f"{message} offset: None")

    def set_offset_series(self, offset_series):
        """
        Parameters
        ----------
        offset_series: pandas.Series
            series with zone_id index and skim array offset values (can map many zone_ids to skim array index)
        """
        assert isinstance(offset_series, pd.Series)
        self.offset_series = offset_series
        self.offset_int = None

    def set_offset_list(self, offset_list):
        """
        Convenience method to set offset_series using an integer list the same size as target skim dimension
        with implicit skim index mapping (e.g. an omx mapping as returned by omx_file.mapentries)

        Parameters
        ----------
        offset_list : list of int
        """
        assert isinstance(offset_list, list) or isinstance(offset_list, np.ndarray)

        if isinstance(offset_list, np.ndarray):
            offset_list = list(offset_list)

        # - for performance, check if this is a simple range that can ber represented by an int offset
        first_offset = offset_list[0]
        if offset_list == list(range(first_offset, len(offset_list) + first_offset)):
            offset_int = -1 * first_offset
            self.set_offset_int(offset_int)
        else:
            offset_series = pd.Series(
                data=list(range(len(offset_list))), index=offset_list
            )
            self.set_offset_series(offset_series)

    def set_offset_int(self, offset_int):
        """
        specify int offset which when added to zone_id yields skim array index (e.g. -1 to map 1-based to 0-based)

        Parameters
        ----------
        offset_int : int
        """
        # should be some duck subtype of integer (but might be, say, numpy.int64)
        assert int(offset_int) == offset_int

        self.offset_int = int(offset_int)
        self.offset_series = None

    def map(self, zone_ids):
        """
        map zone_ids to skim indexes

        Parameters
        ----------
        zone_ids : list-like (numpy.ndarray, pandas.Int64Index, or pandas.Series)

        Returns
        -------
        offsets : numpy array of int
        """

        if self.offset_series is not None:
            assert self.offset_int is None
            assert isinstance(self.offset_series, pd.Series)

            # FIXME - turns out it is faster to use series.map if zone_ids is a series
            # offsets = quick_loc_series(zone_ids, self.offset_series).fillna(NOT_IN_SKIM_ZONE_ID).astype(int)

            if isinstance(zone_ids, np.ndarray):
                zone_ids = pd.Series(zone_ids)
            offsets = (
                zone_ids.map(self.offset_series, na_action="ignore")
                .fillna(NOT_IN_SKIM_ZONE_ID)
                .astype(int)
            )

        elif self.offset_int:
            assert self.offset_series is None
            # apply integer offset, but map NOT_IN_SKIM_ZONE_ID to self
            offsets = np.where(
                zone_ids == NOT_IN_SKIM_ZONE_ID,
                NOT_IN_SKIM_ZONE_ID,
                zone_ids + self.offset_int,
            )
        else:
            offsets = zone_ids

        return offsets


class SkimDict(object):
    """
    A SkimDict object is a wrapper around a dict of multiple skim objects,
    where each object is identified by a key.

    Note that keys are either strings or tuples of two strings (to support stacking of skims.)
    """

    def __init__(self, skim_tag, skim_info, skim_data):

        logger.info(f"SkimDict init {skim_tag}")

        self.skim_tag = skim_tag
        self.skim_info = skim_info
        self.usage = set()  # track keys of skims looked up

        self.offset_mapper = (
            self._offset_mapper()
        )  # (in function so subclass can override)

        self.omx_shape = skim_info.omx_shape
        self.skim_data = skim_data
        self.dtype = np.dtype(
            skim_info.dtype_name
        )  # so we can coerce if we have missing values

        # - skim_dim3 dict maps key1 to dict of key2 absolute offsets into block
        # DRV_COM_WLK_BOARDS: {'MD': 4, 'AM': 3, 'PM': 5}, ...
        self.skim_dim3 = {}

        for skim_key, offset in skim_info.block_offsets.items():
            if isinstance(skim_key, tuple):
                key1, key2 = skim_key
                self.skim_dim3.setdefault(key1, {})[key2] = offset
        logger.info(
            f"SkimDict.build_3d_skim_block_offset_table registered {len(self.skim_dim3)} 3d keys"
        )

    def _offset_mapper(self):
        """
        Return an OffsetMapper to set self.offset_mapper for use with skims
        This allows subclasses (e.g. MazSkimDict) to 'tweak' the parent offset mapper.

        Returns
        -------
        OffsetMapper
        """
        offset_mapper = OffsetMapper()
        if self.skim_info.offset_map is not None:
            offset_mapper.set_offset_list(offset_list=self.skim_info.offset_map)
        else:
            # assume this is a one-based skim map
            offset_mapper.set_offset_int(-1)

        return offset_mapper

    @property
    def zone_ids(self):
        """
        Return list of zone_ids we grok in skim index order

        Returns
        -------
        ndarray of int domain zone_ids
        """

        if self.offset_mapper.offset_series is not None:
            ids = self.offset_mapper.offset_series.index.values
        else:
            ids = np.array(range(self.omx_shape[0])) - self.offset_mapper.offset_int
        return ids

    def get_skim_usage(self):
        """
        return set of keys of skims looked up. e.g. {'DIST', 'SOV'}

        Returns
        -------
        set:
        """
        return self.usage

    def _lookup(self, orig, dest, block_offsets):
        """
        Return list of skim values of skims(s) at orig/dest for the skim(s) at block_offset in skim_data

        Supplying a single int block_offset makes the lookup 2-D
        Supplying a list of block_offsets (same length as orig and dest lists) allows 3D lookup

        Parameters
        ----------
        orig: list of orig zone_ids
        dest: list of dest zone_ids
        block_offsets: int or list of dim3 blockoffsets for the od pairs

        Returns
        -------
        Numpy.ndarray: list of skim values for od pairs
        """

        # fixme - remove?
        assert not (np.isnan(orig) | np.isnan(dest)).any()

        # only working with numpy in here
        orig = np.asanyarray(orig).astype(int)
        dest = np.asanyarray(dest).astype(int)

        mapped_orig = self.offset_mapper.map(orig)
        mapped_dest = self.offset_mapper.map(dest)
        if ROW_MAJOR_LAYOUT:
            result = self.skim_data[block_offsets, mapped_orig, mapped_dest]
        else:
            result = self.skim_data[mapped_orig, mapped_dest, block_offsets]

        # FIXME - should return nan if not in skim (negative indices wrap around)
        in_skim = (
            (mapped_orig >= 0)
            & (mapped_orig < self.omx_shape[0])
            & (mapped_dest >= 0)
            & (mapped_dest < self.omx_shape[1])
        )

        # if not ((in_skim | (orig == NOT_IN_SKIM_ZONE_ID) | (dest == NOT_IN_SKIM_ZONE_ID)).all()):
        #     print(f"orig\n{orig}")
        #     print(f"dest\n{dest}")
        #     print(f"in_skim\n{in_skim}")

        # check for bad indexes (other than NOT_IN_SKIM_ZONE_ID)
        assert (
            in_skim | (orig == NOT_IN_SKIM_ZONE_ID) | (dest == NOT_IN_SKIM_ZONE_ID)
        ).all(), f"{(~in_skim).sum()} od pairs not in skim"

        if not in_skim.all():
            result = np.where(in_skim, result, NOT_IN_SKIM_NAN).astype(self.dtype)

        return result

    def lookup(self, orig, dest, key):
        """
        Return list of skim values of skims(s) at orig/dest in skim with the specified key (e.g. 'DIST')

        Parameters
        ----------
        orig: list of orig zone_ids
        dest: list of dest zone_ids
        key: str

        Returns
        -------
        Numpy.ndarray: list of skim values for od pairs
        """

        self.usage.add(key)

        block_offset = self.skim_info.block_offsets.get(key)
        assert block_offset is not None, f"SkimDict lookup key '{key}' not in skims"

        try:
            result = self._lookup(orig, dest, block_offset)
        except Exception as err:
            logger.error("SkimDict lookup error: %s: %s", type(err).__name__, str(err))
            logger.error(f"key {key}")
            logger.error(f"orig max {orig.max()} min {orig.min()}")
            logger.error(f"dest max {dest.max()} min {dest.min()}")
            raise err

        return result

    def lookup_3d(self, orig, dest, dim3, key):
        """
        3D lookup of skim values of skims(s) at orig/dest for stacked skims indexed by dim3 selector

        The idea is that skims may be stacked in groups with a base key and a dim3 key (usually a time of day key)

        On import (from omx) skims stacks are represented by base and dim3 keys seperated by a double_underscore

        e.g. DRV_COM_WLK_BOARDS__AM indicates base skim key DRV_COM_WLK_BOARDS with a time of day (dim3) of 'AM'

        Since all the skimsa re stored in a single contiguous 3D array, we can use the dim3 key as a third index
        and thus rapidly get skim values for a list of (orig, dest, tod) tuples using index arrays ('fancy indexing')

        Parameters
        ----------
        orig: list of orig zone_ids
        dest: list of dest zone_ids
        block_offsets: list with one dim3 key for each orig/dest pair

        Returns
        -------
        Numpy.ndarray: list of skim values
        """

        self.usage.add(key)  # should we keep usage stats by (key, dim3)?

        assert key in self.skim_dim3, f"3d skim key {key} not in skims."

        # map dim3 to block_offsets
        skim_keys_to_indexes = self.skim_dim3[key]

        # skim_indexes = dim3.map(skim_keys_to_indexes).astype('int')
        try:
            block_offsets = np.vectorize(skim_keys_to_indexes.get)(
                dim3
            )  # this should be faster than map
            result = self._lookup(orig, dest, block_offsets)
        except Exception as err:
            logger.error(
                "SkimDict lookup_3d error: %s: %s", type(err).__name__, str(err)
            )
            logger.error(f"key {key}")
            logger.error(f"orig max {orig.max()} min {orig.min()}")
            logger.error(f"dest max {dest.max()} min {dest.min()}")
            logger.error(f"skim_keys_to_indexes: {skim_keys_to_indexes}")
            logger.error(f"dim3 {np.unique(dim3)}")
            logger.error(f"dim3 block_offsets {np.unique(block_offsets)}")
            raise err

        return result

    def wrap(self, orig_key, dest_key):
        """
        return a SkimWrapper for self
        """
        return SkimWrapper(self, orig_key, dest_key)

    def wrap_3d(self, orig_key, dest_key, dim3_key):
        """
        return a SkimWrapper for self
        """
        return Skim3dWrapper(self, orig_key, dest_key, dim3_key)


class SkimWrapper(object):

    """
    A SkimWrapper object is an access wrapper around a SkimDict of multiple skim objects,
    where each object is identified by a key.

    This is just a way to simplify expression files by hiding the and orig, dest arguments
    when the orig and dest vectors are in a dataframe with known column names (specified at init time)
    The dataframe is identified by set_df because it may not be available (e.g. due to chunking)
    at the time the SkimWrapper is instantiated.

    When the user calls skims[key], key is an identifier for which skim
    to use, and the object automatically looks up impedances of that skim
    using the specified orig_key column in df as the origin and
    the dest_key column in df as the destination.  In this way, the user
    does not do the O-D lookup by hand and only specifies which skim to use
    for this lookup.  This is the only purpose of this object: to
    abstract away the O-D lookup and use skims by specifying which skim
    to use in the expressions.

    Note that keys are either strings or tuples of two strings (to support stacking of skims.)
    """

    def __init__(self, skim_dict, orig_key, dest_key):
        """

        Parameters
        ----------
        skim_dict: SkimDict

        orig_key: str
            name of column in dataframe to use as implicit orig for lookups
        dest_key: str
            name of column in dataframe to use as implicit dest for lookups
        """
        self.skim_dict = skim_dict
        self.orig_key = orig_key
        self.dest_key = dest_key
        self.df = None

    def set_df(self, df):
        """
        Set the dataframe

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the origin and destination ids

        Returns
        -------
        self (to facilitiate chaining)
        """
        assert (
            self.orig_key in df
        ), f"orig_key '{self.orig_key}' not in df columns: {list(df.columns)}"
        assert (
            self.dest_key in df
        ), f"dest_key '{self.dest_key}' not in df columns: {list(df.columns)}"
        self.df = df
        return self

    def lookup(self, key, reverse=False):
        """
        Generally not called by the user - use __getitem__ instead

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        od : bool (optional)
            od=True means lookup standard origin-destination skim value
            od=False means lookup destination-origin skim value

        Returns
        -------
        impedances: pd.Series
            A Series of impedances which are elements of the Skim object and
            with the same index as df
        """

        assert self.df is not None, "Call set_df first"

        if reverse:
            s = self.skim_dict.lookup(
                self.df[self.dest_key], self.df[self.orig_key], key
            )
        else:
            s = self.skim_dict.lookup(
                self.df[self.orig_key], self.df[self.dest_key], key
            )

        return pd.Series(s, index=self.df.index)

    def reverse(self, key):
        """
        return skim value in reverse (d-o) direction
        """
        return self.lookup(key, reverse=True)

    def max(self, key):
        """
        return max skim value in either o-d or d-o direction
        """
        assert self.df is not None, "Call set_df first"

        s = np.maximum(
            self.skim_dict.lookup(self.df[self.dest_key], self.df[self.orig_key], key),
            self.skim_dict.lookup(self.df[self.orig_key], self.df[self.dest_key], key),
        )

        return pd.Series(s, index=self.df.index)

    def __getitem__(self, key):
        """
        Get the lookup for an available skim object (df and orig/dest and column names implicit)

        Parameters
        ----------
        key : hashable
             The key (identifier) for the skim object

        Returns
        -------
        impedances: pd.Series with the same index as df
            A Series of impedances values from the single Skim with specified key, indexed byt orig/dest pair
        """

        return self.lookup(key)


class Skim3dWrapper(object):
    """

    This works the same as a SkimWrapper above, except the third dim3 is also supplied,
    and a 3D lookup is performed using orig, dest, and dim3.

    Parameters
    ----------
    skims: Skims
        This is the Skims object to wrap
    dim3_key : str
        This identifies the column in the dataframe which is used to
        select among Skim object using the SECOND item in each tuple (see
        above for a more complete description)
    """

    def __init__(self, skim_dict, orig_key, dest_key, dim3_key):
        """

        Parameters
        ----------
        skim_dict: SkimDict

        orig_key: str
            name of column of zone_ids in dataframe to use as implicit orig for lookups
        dest_key: str
            name of column of zone_ids  in dataframe to use as implicit dest for lookups
        dim3_key: str
            name of column of dim3 keys in dataframe to use as implicit third dim3 key for 3D lookups
            e.g. string column with time_of_day keys (such as 'AM', 'MD', 'PM', etc.)
        """
        self.skim_dict = skim_dict

        self.orig_key = orig_key
        self.dest_key = dest_key
        self.dim3_key = dim3_key
        self.df = None

    def set_df(self, df):
        """
        Set the dataframe

        Parameters
        ----------
        df : DataFrame
            The dataframe which contains the orig, dest, and dim3 values

        Returns
        -------
        self (to facilitiate chaining)
        """
        assert (
            self.orig_key in df
        ), f"orig_key '{self.orig_key}' not in df columns: {list(df.columns)}"
        assert (
            self.dest_key in df
        ), f"dest_key '{self.dest_key}' not in df columns: {list(df.columns)}"
        assert (
            self.dim3_key in df
        ), f"dim3_key '{self.dim3_key}' not in df columns: {list(df.columns)}"
        self.df = df
        return self

    def __getitem__(self, key):
        """
        Get the lookup for an available skim object (df and orig/dest/dim3 and column names implicit)

        Parameters
        ----------
        key : hashable
             The key (identifier) for this skim object

        Returns
        -------
        impedances: pd.Series with the same index as df
            A Series of impedances values from the set of skims with specified base key, indexed by orig/dest/dim3
        """
        assert self.df is not None, "Call set_df first"
        orig = self.df[self.orig_key].astype("int")
        dest = self.df[self.dest_key].astype("int")
        dim3 = self.df[self.dim3_key]

        skim_values = self.skim_dict.lookup_3d(orig, dest, dim3, key)

        return pd.Series(skim_values, self.df.index)


class MazSkimDict(SkimDict):
    """
    MazSkimDict provides a facade that allows skim-like lookup by maz orig,dest zone_id
    when there are often too many maz zones to create maz skims.

    Dependencies: network_los.load_data must have already loaded: taz skim_dict, maz_to_maz_df, and maz_taz_df

    It performs lookups from a sparse list of maz-maz od pairs on selected attributes (e.g. WALKDIST)
    where accuracy for nearby od pairs is critical. And is backed by a fallback taz skim dict
    to return values of for more distant pairs (or for skims that are not attributes in the maz-maz table.)
    """

    def __init__(self, skim_tag, network_los, taz_skim_dict):
        """
        we need network_los because we have dependencies on network_los.load_data (e.g. maz_to_maz_df, maz_taz_df,
        and the fallback taz skim_dict)

        We require taz_skim_dict as an explicit parameter to emphasize that we are piggybacking on taz_skim_dict's
        preexisting skim_data and skim_info, rather than instantiating duplicate copies thereof.

        Note, however, that we override _offset_mapper (called by super.__init__) to create our own
        custom self.offset_mapper that maps directly from MAZ zone_ids to TAZ skim array indexes

        Parameters
        ----------
        skim_tag: str
        network_los: Network_LOS
        taz_skim_dict: SkimDict
        """

        self.network_los = network_los

        super().__init__(skim_tag, taz_skim_dict.skim_info, taz_skim_dict.skim_data)
        assert (
            self.offset_mapper is not None
        )  # should have been set with _init_offset_mapper

        self.dtype = np.dtype(self.skim_info.dtype_name)
        self.base_keys = taz_skim_dict.skim_info.base_keys
        self.sparse_keys = list(
            set(network_los.maz_to_maz_df.columns) - {"OMAZ", "DMAZ"}
        )
        self.sparse_key_usage = set()

    def _offset_mapper(self):
        """
        return an OffsetMapper to map maz zone_ids to taz skim indexes
        Specifically, an offset_series with MAZ zone_id index and TAZ skim array offset values

        This is called by super().__init__ AFTER

        Returns
        -------
        OffsetMapper
        """

        # start with a series with MAZ zone_id index and TAZ zone id values
        maz_to_taz = (
            self.network_los.maz_taz_df[["MAZ", "TAZ"]]
            .set_index("MAZ")
            .sort_values(by="TAZ")
            .TAZ
        )

        # use taz offset_mapper to create series mapping directly from MAZ to TAZ skim index
        taz_offset_mapper = super()._offset_mapper()
        maz_to_skim_offset = taz_offset_mapper.map(maz_to_taz)

        if isinstance(maz_to_skim_offset, np.ndarray):
            maz_to_skim_offset = pd.Series(maz_to_skim_offset, maz_to_taz.index)  # bug

        # MAZ
        # 19062    330 <- The TAZ would be, say, 331, and the offset is 330
        # 8429     330
        # 9859     331

        assert isinstance(maz_to_skim_offset, np.ndarray) or isinstance(
            maz_to_skim_offset, pd.Series
        )
        if isinstance(maz_to_skim_offset, pd.Series):
            offset_mapper = OffsetMapper(offset_series=maz_to_skim_offset)
        elif isinstance(maz_to_skim_offset, np.ndarray):
            offset_mapper = OffsetMapper(offset_list=maz_to_skim_offset)

        return offset_mapper

    def get_skim_usage(self):
        return self.sparse_key_usage.union(self.usage)

    def sparse_lookup(self, orig, dest, key):
        """
        Get impedence values for a set of origin, destination pairs.

        Parameters
        ----------
        orig : 1D array
        dest : 1D array
        key : str
            skim key

        Returns
        -------
        values : numpy 1D array
        """

        self.sparse_key_usage.add(key)

        max_blend_distance = self.network_los.max_blend_distance.get(key, 0)

        if max_blend_distance == 0:
            blend_distance_skim_name = None
        else:
            blend_distance_skim_name = self.network_los.blend_distance_skim_name

        # fixme - remove?
        assert not (np.isnan(orig) | np.isnan(dest)).any()

        # we want values from mazpairs, where we have them
        values = self.network_los.get_mazpairs(orig, dest, key)

        is_nan = np.isnan(values)

        if max_blend_distance > 0:

            # print(f"{is_nan.sum()} nans out of {len(is_nan)} for key '{self.key}")
            # print(f"blend_distance_skim_name {self.blend_distance_skim_name}")

            backstop_values = super().lookup(orig, dest, key)

            # get distance skim if a different key was specified by blend_distance_skim_name
            if blend_distance_skim_name != key:
                distance = self.network_los.get_mazpairs(
                    orig, dest, blend_distance_skim_name
                )
            else:
                distance = values

            # for distances less than max_blend_distance, we blend maz-maz and skim backstop values
            # shorter distances have less fractional backstop, and more maz-maz
            # beyond max_blend_distance, just use the skim values
            backstop_fractions = np.minimum(distance / max_blend_distance, 1)

            values = np.where(
                is_nan,
                backstop_values,
                backstop_fractions * backstop_values
                + (1 - backstop_fractions) * values,
            )

        elif is_nan.any():
            # print(f"{is_nan.sum()} nans out of {len(is_nan)} for key '{self.key}")

            if key in self.base_keys:

                # replace nan values using simple backstop without blending
                backstop_values = super().lookup(orig, dest, key)
                values = np.where(is_nan, backstop_values, values)
            else:
                # FIXME - if no backstop skim, then return 0 (which conventionally means "not available")
                logger.warning(
                    "No backstop skims found for {0}, so setting Nulls to 0. Make sure "
                    "mode availability flags are set to > 0"
                )
                values = np.where(is_nan, 0, values)

        # want to return same type as backstop skim
        values = values.astype(self.dtype)

        return values

    def lookup(self, orig, dest, key):
        """
        Return list of skim values of skims(s) at orig/dest in skim with the specified key (e.g. 'DIST')

        Look up in sparse table (backed by taz skims) if key is a sparse_key, otherwise look up in taz skims
        For taz skim lookups, the offset_mapper will convert maz zone_ids directly to taz skim indexes.

        Parameters
        ----------
        orig: list of orig zone_ids
        dest: list of dest zone_ids
        key: str

        Returns
        -------
        Numpy.ndarray: list of skim values for od pairs
        """

        if key in self.sparse_keys:
            # logger.debug(f"MazSkimDict using SparseSkimDict for key '{key}'")
            values = self.sparse_lookup(orig, dest, key)
        else:
            values = super().lookup(orig, dest, key)

        return values


class DataFrameMatrix(object):
    """
    Utility class to allow a pandas dataframe to be treated like a 2-D array,
    indexed by rowid, colname

    For use in vectorized expressions where the desired values depend on both a row column selector
    e.g. size_terms.get(df.dest_taz, df.purpose)

    ::

      df = pd.DataFrame({'a': [1,2,3,4,5], 'b': [10,20,30,40,50]}, index=[100,101,102,103,104])

      dfm = DataFrameMatrix(df)

      dfm.get(row_ids=[100,100,103], col_ids=['a', 'b', 'a'])

      returns [1, 10,  4]

    """

    def __init__(self, df):
        """

        Parameters
        ----------
        df - pandas dataframe of uniform type
        """

        self.df = df
        self.data = df.values

        self.offset_mapper = OffsetMapper()
        self.offset_mapper.set_offset_list(list(df.index))

        self.cols_to_indexes = {k: v for v, k in enumerate(df.columns)}

    def get(self, row_ids, col_ids):
        """

        Parameters
        ----------
        row_ids - list of row_ids (df index values)
        col_ids - list of column names, one per row_id,
                  specifying column from which the value for that row should be retrieved

        Returns
        -------

        series with one row per row_id, with the value from the column specified in col_ids

        """
        # col_indexes = segments.map(self.cols_to_indexes).astype('int')
        # this should be faster than map
        col_indexes = np.vectorize(self.cols_to_indexes.get)(col_ids)

        row_indexes = self.offset_mapper.map(np.asanyarray(row_ids))

        not_in_skim = row_indexes == NOT_IN_SKIM_ZONE_ID
        if not_in_skim.any():
            logger.warning(
                f"DataFrameMatrix: {not_in_skim.sum()} row_ids of {len(row_ids)} not in skim."
            )
            not_in_skim = not_in_skim.values
            logger.warning(f"row_ids: {row_ids[not_in_skim]}")
            logger.warning(f"col_ids: {col_ids[not_in_skim]}")
            raise RuntimeError(
                f"DataFrameMatrix: {not_in_skim.sum()} row_ids of {len(row_ids)} not in skim."
            )

        assert (row_indexes >= 0).all(), f"{row_indexes}"

        result = self.data[row_indexes, col_indexes]

        # FIXME - if ids (or col_ids?) is a series, return series with same index?
        if isinstance(row_ids, pd.Series):
            result = pd.Series(result, index=row_ids.index)

        return result



# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import range
from math import ceil

import numpy as np
import pandas as pd

from . import chunk, interaction_simulate, logit, pipeline, tracing
from .simulate import set_skim_wrapper_targets

logger = logging.getLogger(__name__)

DUMP = False


def make_sample_choices(
    choosers,
    probs,
    alternatives,
    sample_size,
    alternative_count,
    alt_col_name,
    allow_zero_probs,
    trace_label,
):
    """

    Parameters
    ----------
    choosers
    probs : pandas DataFrame
        one row per chooser and one column per alternative
    alternatives
        dataframe with index containing alt ids
    sample_size : int
        number of samples/choices to make
    alternative_count
    alt_col_name : str
    trace_label

    Returns
    -------

    """

    assert isinstance(probs, pd.DataFrame)
    assert probs.shape == (len(choosers), alternative_count)

    assert isinstance(alternatives, pd.DataFrame)
    assert len(alternatives) == alternative_count

    if allow_zero_probs:
        zero_probs = probs.sum(axis=1) == 0
        if zero_probs.all():
            return pd.DataFrame(
                columns=[alt_col_name, "rand", "prob", choosers.index.name]
            )
        if zero_probs.any():
            # remove from sample
            probs = probs[~zero_probs]
            choosers = choosers[~zero_probs]

    cum_probs_array = probs.values.cumsum(axis=1)
    chunk.log_df(trace_label, "cum_probs_array", cum_probs_array)

    # alt probs in convenient layout to return prob of chose alternative
    # (same layout as cum_probs_arr)
    alt_probs_array = probs.values.flatten()
    chunk.log_df(trace_label, "alt_probs_array", alt_probs_array)

    # get sample_size rands for each chooser
    rands = pipeline.get_rn_generator().random_for_df(probs, n=sample_size)

    # transform as we iterate over alternatives
    # reshape so rands[i] is in broadcastable (2-D) shape for cum_probs_arr
    # i.e rands[i] is a 2-D array of one alt choice rand for each chooser
    rands = rands.T.reshape(sample_size, -1, 1)
    chunk.log_df(trace_label, "rands", rands)

    # the alternative value chosen
    # WHY SHOULD CHOICES COL HAVE TO BE TYPE INT???
    # choices_array = np.empty([sample_size, len(choosers)]).astype(int)
    choices_array = np.empty([sample_size, len(choosers)]).astype(
        alternatives.index.dtype
    )
    # chunk log these later after we populate them...

    # the probability of the chosen alternative
    choice_probs_array = np.empty([sample_size, len(choosers)])
    # chunk log these later after we populate them...

    alts = np.tile(alternatives.index.values, len(choosers))
    chunk.log_df(trace_label, "alts", alts)

    # FIXME - do this all at once rather than iterate?
    for i in range(sample_size):

        # FIXME - do this in numpy, not pandas?

        # rands for this alt in broadcastable shape
        r = rands[i]

        # position of first occurrence of positive value
        positions = np.argmax(cum_probs_array > r, axis=1)

        # FIXME - leave positions as numpy array, not pandas series?
        # positions is series with the chosen alternative represented as a column index in probs
        # which is an integer between zero and num alternatives in the alternative sample
        positions = pd.Series(positions, index=probs.index)

        # need to get from an integer offset into the alternative sample to the alternative index
        # that is, we want the index value of the row that is offset by <position> rows into the
        # tranche of this choosers alternatives created by cross join of alternatives and choosers

        # offsets is the offset into model_design df of first row of chooser alternatives
        offsets = np.arange(len(positions)) * alternative_count

        # choices and choice_probs have one element per chooser and is in same order as choosers
        choices_array[i] = np.take(alts, positions + offsets)
        choice_probs_array[i] = np.take(alt_probs_array, positions + offsets)

        del positions
        del offsets

    chunk.log_df(trace_label, "choices_array", choices_array)
    chunk.log_df(trace_label, "choice_probs_array", choice_probs_array)

    del alts
    chunk.log_df(trace_label, "alts", None)
    del cum_probs_array
    chunk.log_df(trace_label, "cum_probs_array", None)
    del alt_probs_array
    chunk.log_df(trace_label, "alt_probs_array", None)

    # explode to one row per chooser.index, alt_zone_id
    choices_df = pd.DataFrame(
        {
            alt_col_name: choices_array.flatten(order="F"),
            "rand": rands.flatten(order="F"),
            "prob": choice_probs_array.flatten(order="F"),
            choosers.index.name: np.repeat(np.asanyarray(choosers.index), sample_size),
        }
    )

    chunk.log_df(trace_label, "choices_df", choices_df)

    del choices_array
    chunk.log_df(trace_label, "choices_array", None)
    del rands
    chunk.log_df(trace_label, "rands", None)
    del choice_probs_array
    chunk.log_df(trace_label, "choice_probs_array", None)

    # handing this off to caller
    chunk.log_df(trace_label, "choices_df", None)

    return choices_df


def _interaction_sample(
    choosers,
    alternatives,
    spec,
    sample_size,
    alt_col_name,
    allow_zero_probs=False,
    log_alt_losers=False,
    skims=None,
    locals_d=None,
    trace_label=None,
):
    """
    Run a MNL simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    Parameters are same as for public function interaction_sa,ple

    spec : dataframe
        one row per spec expression and one col with utility coefficient

    interaction_df : dataframe
        cross join (cartesian product) of choosers with alternatives
        combines columns of choosers and alternatives
        len(df) == len(choosers) * len(alternatives)
        index values (non-unique) are index values from alternatives df

    interaction_utilities : dataframe
        the utility of each alternative is sum of the partial utilities determined by the
        various spec expressions and their corresponding coefficients
        yielding a dataframe  with len(interaction_df) rows and one utility column
        having the same index as interaction_df (non-unique values from alternatives df)

    utilities : dataframe
        dot product of model_design.dot(spec)
        yields utility value for element in the cross product of choosers and alternatives
        this is then reshaped as a dataframe with one row per chooser and one column per alternative

    probs : dataframe
        utilities exponentiated and converted to probabilities
        same shape as utilities, one row per chooser and one column per alternative

    positions : series
        choices among alternatives with the chosen alternative represented
        as the integer index of the selected alternative column in probs

    choices : series
        series with the alternative chosen for each chooser
        the index is same as choosers
        and the series value is the alternative df index of chosen alternative

    Returns
    -------
    choices_df : pandas.DataFrame

        A DataFrame where index should match the index of the choosers DataFrame
        and columns alt_col_name, prob, rand, pick_count

        prob: float
            the probability of the chosen alternative
        rand: float
            the rand that did the choosing
        pick_count : int
            number of duplicate picks for chooser, alt
    """

    have_trace_targets = tracing.has_trace_targets(choosers)
    num_choosers = len(choosers.index)

    assert num_choosers > 0

    if have_trace_targets:
        tracing.trace_df(choosers, tracing.extend_trace_label(trace_label, "choosers"))
        tracing.trace_df(
            alternatives,
            tracing.extend_trace_label(trace_label, "alternatives"),
            slicer="NONE",
            transpose=False,
        )

    if len(spec.columns) > 1:
        raise RuntimeError("spec must have only one column")

    # if using skims, copy index into the dataframe, so it will be
    # available as the "destination" for set_skim_wrapper_targets
    if skims is not None and alternatives.index.name not in alternatives:
        alternatives = alternatives.copy()
        alternatives[alternatives.index.name] = alternatives.index

    chooser_index_id = interaction_simulate.ALT_CHOOSER_ID if log_alt_losers else None

    # - cross join choosers and alternatives (cartesian product)
    # for every chooser, there will be a row for each alternative
    # index values (non-unique) are from alternatives df
    alternative_count = alternatives.shape[0]
    interaction_df = logit.interaction_dataset(
        choosers,
        alternatives,
        sample_size=alternative_count,
        chooser_index_id=chooser_index_id,
    )

    chunk.log_df(trace_label, "interaction_df", interaction_df)

    assert alternative_count == len(interaction_df.index) / len(choosers.index)

    if skims is not None:
        set_skim_wrapper_targets(interaction_df, skims)

    # evaluate expressions from the spec multiply by coefficients and sum
    # spec is df with one row per spec expression and one col with utility coefficient
    # column names of interaction_df match spec index values
    # utilities has utility value for element in the cross product of choosers and alternatives
    # interaction_utilities is a df with one utility column and one row per row in interaction_df
    if have_trace_targets:
        trace_rows, trace_ids = tracing.interaction_trace_rows(
            interaction_df, choosers, alternative_count
        )

        tracing.trace_df(
            interaction_df[trace_rows],
            tracing.extend_trace_label(trace_label, "interaction_df"),
            slicer="NONE",
            transpose=False,
        )
    else:
        trace_rows = trace_ids = None

    # interaction_utilities is a df with one utility column and one row per interaction_df row
    (
        interaction_utilities,
        trace_eval_results,
    ) = interaction_simulate.eval_interaction_utilities(
        spec,
        interaction_df,
        locals_d,
        trace_label,
        trace_rows,
        estimator=None,
        log_alt_losers=log_alt_losers,
    )
    chunk.log_df(trace_label, "interaction_utilities", interaction_utilities)

    # ########### HWM - high water mark (point of max observed memory usage)

    del interaction_df
    chunk.log_df(trace_label, "interaction_df", None)

    if have_trace_targets:
        tracing.trace_interaction_eval_results(
            trace_eval_results,
            trace_ids,
            tracing.extend_trace_label(trace_label, "eval"),
        )

        tracing.trace_df(
            interaction_utilities[trace_rows],
            tracing.extend_trace_label(trace_label, "interaction_utilities"),
            slicer="NONE",
            transpose=False,
        )

    tracing.dump_df(DUMP, interaction_utilities, trace_label, "interaction_utilities")

    # reshape utilities (one utility column and one row per row in interaction_utilities)
    # to a dataframe with one row per chooser and one column per alternative
    utilities = pd.DataFrame(
        interaction_utilities.values.reshape(len(choosers), alternative_count),
        index=choosers.index,
    )
    chunk.log_df(trace_label, "utilities", utilities)

    del interaction_utilities
    chunk.log_df(trace_label, "interaction_utilities", None)

    if have_trace_targets:
        tracing.trace_df(
            utilities,
            tracing.extend_trace_label(trace_label, "utils"),
            column_labels=["alternative", "utility"],
        )

    tracing.dump_df(DUMP, utilities, trace_label, "utilities")

    # convert to probabilities (utilities exponentiated and normalized to probs)
    # probs is same shape as utilities, one row per chooser and one column for alternative
    probs = logit.utils_to_probs(
        utilities,
        allow_zero_probs=allow_zero_probs,
        trace_label=trace_label,
        trace_choosers=choosers,
    )
    chunk.log_df(trace_label, "probs", probs)

    del utilities
    chunk.log_df(trace_label, "utilities", None)

    if have_trace_targets:
        tracing.trace_df(
            probs,
            tracing.extend_trace_label(trace_label, "probs"),
            column_labels=["alternative", "probability"],
        )

    if sample_size == 0:
        # FIXME return full alternative set rather than sample
        logger.info(
            "Estimation mode for %s using unsampled alternatives" % (trace_label,)
        )

        index_name = probs.index.name
        choices_df = (
            pd.melt(probs.reset_index(), id_vars=[index_name])
            .sort_values(by=index_name, kind="mergesort")
            .set_index(index_name)
            .rename(columns={"value": "prob"})
            .drop(columns="variable")
        )

        choices_df["pick_count"] = 1
        choices_df.insert(
            0, alt_col_name, np.tile(alternatives.index.values, len(choosers.index))
        )

        return choices_df
    else:
        choices_df = make_sample_choices(
            choosers,
            probs,
            alternatives,
            sample_size,
            alternative_count,
            alt_col_name,
            allow_zero_probs=allow_zero_probs,
            trace_label=trace_label,
        )

    chunk.log_df(trace_label, "choices_df", choices_df)

    del probs
    chunk.log_df(trace_label, "probs", None)

    # pick_count and pick_dup
    # pick_count is number of duplicate picks
    # pick_dup flag is True for all but first of duplicates
    pick_group = choices_df.groupby([choosers.index.name, alt_col_name])

    # number each item in each group from 0 to the length of that group - 1.
    choices_df["pick_count"] = pick_group.cumcount(ascending=True)
    # flag duplicate rows after first
    choices_df["pick_dup"] = choices_df["pick_count"] > 0
    # add reverse cumcount to get total pick_count (conveniently faster than groupby.count + merge)
    choices_df["pick_count"] += pick_group.cumcount(ascending=False) + 1

    # drop the duplicates
    choices_df = choices_df[~choices_df["pick_dup"]]
    del choices_df["pick_dup"]
    chunk.log_df(trace_label, "choices_df", choices_df)

    # set index after groupby so we can trace on it
    choices_df.set_index(choosers.index.name, inplace=True)

    tracing.dump_df(DUMP, choices_df, trace_label, "choices_df")

    if have_trace_targets:
        tracing.trace_df(
            choices_df,
            tracing.extend_trace_label(trace_label, "sampled_alternatives"),
            transpose=False,
            column_labels=["sample_alt", "alternative"],
        )

    # don't need this after tracing
    del choices_df["rand"]
    chunk.log_df(trace_label, "choices_df", choices_df)

    # - NARROW
    choices_df["prob"] = choices_df["prob"].astype(np.float32)
    assert (choices_df["pick_count"].max() < 4294967295) or (choices_df.empty)
    choices_df["pick_count"] = choices_df["pick_count"].astype(np.uint32)

    return choices_df


def interaction_sample(
    choosers,
    alternatives,
    spec,
    sample_size,
    alt_col_name,
    allow_zero_probs=False,
    log_alt_losers=False,
    skims=None,
    locals_d=None,
    chunk_size=0,
    chunk_tag=None,
    trace_label=None,
):

    """
    Run a simulation in the situation in which alternatives must
    be merged with choosers because there are interaction terms or
    because alternatives are being sampled.

    optionally (if chunk_size > 0) iterates over choosers in chunk_size chunks

    Parameters
    ----------
    choosers : pandas.DataFrame
        DataFrame of choosers
    alternatives : pandas.DataFrame
        DataFrame of alternatives - will be merged with choosers and sampled
    spec : pandas.DataFrame
        A Pandas DataFrame that gives the specification of the variables to
        compute and the coefficients for each variable.
        Variable specifications must be in the table index and the
        table should have only one column of coefficients.
    sample_size : int, optional
        Sample alternatives with sample of given size.  By default is None,
        which does not sample alternatives.
    alt_col_name: str
        name to give the sampled_alternative column
    skims : Skims object
        The skims object is used to contain multiple matrices of
        origin-destination impedances.  Make sure to also add it to the
        locals_d below in order to access it in expressions.  The *only* job
        of this method in regards to skims is to call set_df with the
        dataframe that comes back from interacting choosers with
        alternatives.  See the skims module for more documentation on how
        the skims object is intended to be used.
    locals_d : Dict
        This is a dictionary of local variables that will be the environment
        for an evaluation of an expression that begins with @
    chunk_size : int
        if chunk_size > 0 iterates over choosers in chunk_size chunks
    trace_label: str
        This is the label to be used  for trace log file entries and dump file names
        when household tracing enabled. No tracing occurs if label is empty or None.

    Returns
    -------
    choices_df : pandas.DataFrame

        A DataFrame where index should match the index of the choosers DataFrame
        (except with sample_size rows for each choser row, one row for each alt sample)
        and columns alt_col_name, prob, rand, pick_count

        <alt_col_name>:
            alt identifier from alternatives[<alt_col_name>
        prob: float
            the probability of the chosen alternative
        pick_count : int
            number of duplicate picks for chooser, alt
    """

    trace_label = tracing.extend_trace_label(trace_label, "interaction_sample")
    chunk_tag = chunk_tag or trace_label

    # we return alternatives ordered in (index, alt_col_name)
    # if choosers index is not ordered, it is probably a mistake, since the alts wont line up
    assert alt_col_name is not None
    assert choosers.index.is_monotonic_increasing

    # FIXME - legacy logic - not sure this is needed or even correct?
    sample_size = min(sample_size, len(alternatives.index))

    result_list = []
    for i, chooser_chunk, chunk_trace_label in chunk.adaptive_chunked_choosers(
        choosers, chunk_size, trace_label, chunk_tag
    ):

        choices = _interaction_sample(
            chooser_chunk,
            alternatives,
            spec=spec,
            sample_size=sample_size,
            alt_col_name=alt_col_name,
            allow_zero_probs=allow_zero_probs,
            log_alt_losers=log_alt_losers,
            skims=skims,
            locals_d=locals_d,
            trace_label=chunk_trace_label,
        )

        if choices.shape[0] > 0:
            # might not be any if allow_zero_probs
            result_list.append(choices)

            chunk.log_df(trace_label, f"result_list", result_list)

    # FIXME: this will require 2X RAM
    # if necessary, could append to hdf5 store on disk:
    # http://pandas.pydata.org/pandas-docs/stable/io.html#id2
    if len(result_list) > 1:
        choices = pd.concat(result_list)

    assert allow_zero_probs or (
        len(choosers.index) == len(np.unique(choices.index.values))
    )

    # keep alts in canonical order so choices based on their probs are stable across runs
    choices = choices.sort_values(by=alt_col_name).sort_index(kind="mergesort")

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import object

import numpy as np
import pandas as pd

from . import config, pipeline, tracing

logger = logging.getLogger(__name__)

EXP_UTIL_MIN = 1e-300
EXP_UTIL_MAX = np.inf

PROB_MIN = 0.0
PROB_MAX = 1.0


def report_bad_choices(
    bad_row_map, df, trace_label, msg, trace_choosers=None, raise_error=True
):
    """

    Parameters
    ----------
    bad_row_map
    df : pandas.DataFrame
        utils or probs dataframe
    msg : str
        message describing the type of bad choice that necessitates error being thrown
    trace_choosers : pandas.dataframe
        the choosers df (for interaction_simulate) to facilitate the reporting of hh_id
        because we  can't deduce hh_id from the interaction_dataset which is indexed on index
        values from alternatives df

    Returns
    -------
    raises RuntimeError
    """
    MAX_DUMP = 1000
    MAX_PRINT = 10

    msg_with_count = "%s %s for %s of %s rows" % (
        trace_label,
        msg,
        bad_row_map.sum(),
        len(df),
    )
    logger.warning(msg_with_count)

    df = df[bad_row_map]
    if trace_choosers is None:
        hh_ids, trace_col = tracing.trace_id_for_chooser(df.index, df)
    else:
        hh_ids, trace_col = tracing.trace_id_for_chooser(df.index, trace_choosers)
    df[trace_col] = hh_ids

    if trace_label:
        logger.info("dumping %s" % trace_label)
        tracing.write_csv(df[:MAX_DUMP], file_name=trace_label, transpose=False)

    # log the indexes of the first MAX_DUMP offending rows
    for idx in df.index[:MAX_PRINT].values:

        row_msg = "%s : %s in: %s = %s (hh_id = %s)" % (
            trace_label,
            msg,
            df.index.name,
            idx,
            df[trace_col].loc[idx],
        )

        logger.warning(row_msg)

    if raise_error:
        raise RuntimeError(msg_with_count)


def utils_to_logsums(utils, exponentiated=False, allow_zero_probs=False):
    """
    Convert a table of utilities to logsum series.

    Parameters
    ----------
    utils : pandas.DataFrame
        Rows should be choosers and columns should be alternatives.

    exponentiated : bool
        True if utilities have already been exponentiated

    Returns
    -------
    logsums : pandas.Series
        Will have the same index as `utils`.

    """

    # fixme - conversion to float not needed in either case?
    # utils_arr = utils.values.astype('float')
    utils_arr = utils.values
    if not exponentiated:
        utils_arr = np.exp(utils_arr)

    np.clip(utils_arr, EXP_UTIL_MIN, EXP_UTIL_MAX, out=utils_arr)

    utils_arr = np.where(utils_arr == EXP_UTIL_MIN, 0.0, utils_arr)

    with np.errstate(divide="ignore" if allow_zero_probs else "warn"):
        logsums = np.log(utils_arr.sum(axis=1))

    logsums = pd.Series(logsums, index=utils.index)

    return logsums


def utils_to_probs(
    utils,
    trace_label=None,
    exponentiated=False,
    allow_zero_probs=False,
    trace_choosers=None,
):
    """
    Convert a table of utilities to probabilities.

    Parameters
    ----------
    utils : pandas.DataFrame
        Rows should be choosers and columns should be alternatives.

    trace_label : str
        label for tracing bad utility or probability values

    exponentiated : bool
        True if utilities have already been exponentiated

    allow_zero_probs : bool
        if True value rows in which all utility alts are EXP_UTIL_MIN will result
        in rows in probs to have all zero probability (and not sum to 1.0)
        This is for the benefit of calculating probabilities of nested logit nests

    trace_choosers : pandas.dataframe
        the choosers df (for interaction_simulate) to facilitate the reporting of hh_id
        by report_bad_choices because it can't deduce hh_id from the interaction_dataset
        which is indexed on index values from alternatives df

    Returns
    -------
    probs : pandas.DataFrame
        Will have the same index and columns as `utils`.

    """
    trace_label = tracing.extend_trace_label(trace_label, "utils_to_probs")

    # fixme - conversion to float not needed in either case?
    # utils_arr = utils.values.astype('float')
    utils_arr = utils.values
    if not exponentiated:
        utils_arr = np.exp(utils_arr)

    np.clip(utils_arr, EXP_UTIL_MIN, EXP_UTIL_MAX, out=utils_arr)

    # FIXME
    utils_arr = np.where(utils_arr == EXP_UTIL_MIN, 0.0, utils_arr)

    arr_sum = utils_arr.sum(axis=1)

    zero_probs = arr_sum == 0.0
    if zero_probs.any() and not allow_zero_probs:

        report_bad_choices(
            zero_probs,
            utils,
            trace_label=tracing.extend_trace_label(trace_label, "zero_prob_utils"),
            msg="all probabilities are zero",
            trace_choosers=trace_choosers,
        )

    inf_utils = np.isinf(arr_sum)
    if inf_utils.any():
        report_bad_choices(
            inf_utils,
            utils,
            trace_label=tracing.extend_trace_label(trace_label, "inf_exp_utils"),
            msg="infinite exponentiated utilities",
            trace_choosers=trace_choosers,
        )

    # if allow_zero_probs, this may cause a RuntimeWarning: invalid value encountered in divide
    with np.errstate(
        invalid="ignore" if allow_zero_probs else "warn",
        divide="ignore" if allow_zero_probs else "warn",
    ):
        np.divide(utils_arr, arr_sum.reshape(len(utils_arr), 1), out=utils_arr)

    # if allow_zero_probs, this will cause EXP_UTIL_MIN util rows to have all zero probabilities
    utils_arr[np.isnan(utils_arr)] = PROB_MIN

    np.clip(utils_arr, PROB_MIN, PROB_MAX, out=utils_arr)

    probs = pd.DataFrame(utils_arr, columns=utils.columns, index=utils.index)

    return probs


def make_choices(probs, trace_label=None, trace_choosers=None, allow_bad_probs=False):
    """
    Make choices for each chooser from among a set of alternatives.

    Parameters
    ----------
    probs : pandas.DataFrame
        Rows for choosers and columns for the alternatives from which they
        are choosing. Values are expected to be valid probabilities across
        each row, e.g. they should sum to 1.

    trace_choosers : pandas.dataframe
        the choosers df (for interaction_simulate) to facilitate the reporting of hh_id
        by report_bad_choices because it can't deduce hh_id from the interaction_dataset
        which is indexed on index values from alternatives df

    Returns
    -------
    choices : pandas.Series
        Maps chooser IDs (from `probs` index) to a choice, where the choice
        is an index into the columns of `probs`.

    rands : pandas.Series
        The random numbers used to make the choices (for debugging, tracing)

    """
    trace_label = tracing.extend_trace_label(trace_label, "make_choices")

    # probs should sum to 1 across each row

    BAD_PROB_THRESHOLD = 0.001
    bad_probs = probs.sum(axis=1).sub(
        np.ones(len(probs.index))
    ).abs() > BAD_PROB_THRESHOLD * np.ones(len(probs.index))

    if bad_probs.any() and not allow_bad_probs:

        report_bad_choices(
            bad_probs,
            probs,
            trace_label=tracing.extend_trace_label(trace_label, "bad_probs"),
            msg="probabilities do not add up to 1",
            trace_choosers=trace_choosers,
        )

    rands = pipeline.get_rn_generator().random_for_df(probs)

    probs_arr = probs.values.cumsum(axis=1) - rands

    # rows, cols = np.where(probs_arr > 0)
    # choices = [s.iat[0] for _, s in pd.Series(cols).groupby(rows)]
    choices = np.argmax(probs_arr > 0.0, axis=1)

    choices = pd.Series(choices, index=probs.index)

    rands = pd.Series(np.asanyarray(rands).flatten(), index=probs.index)

    return choices, rands


def interaction_dataset(
    choosers, alternatives, sample_size=None, alt_index_id=None, chooser_index_id=None
):
    """
    Combine choosers and alternatives into one table for the purposes
    of creating interaction variables and/or sampling alternatives.

    Any duplicate column names in choosers table will be renamed with an '_chooser' suffix.

    Parameters
    ----------
    choosers : pandas.DataFrame
    alternatives : pandas.DataFrame
    sample_size : int, optional
        If sampling from alternatives for each chooser, this is
        how many to sample.

    Returns
    -------
    alts_sample : pandas.DataFrame
        Merged choosers and alternatives with data repeated either
        len(alternatives) or `sample_size` times.

    """
    if not choosers.index.is_unique:
        raise RuntimeError(
            "ERROR: choosers index is not unique, " "sample will not work correctly"
        )
    if not alternatives.index.is_unique:
        raise RuntimeError(
            "ERROR: alternatives index is not unique, " "sample will not work correctly"
        )

    numchoosers = len(choosers)
    numalts = len(alternatives)
    sample_size = sample_size or numalts

    # FIXME - is this faster or just dumb?
    alts_idx = np.arange(numalts)

    if sample_size < numalts:
        sample = pipeline.get_rn_generator().choice_for_df(
            choosers, alts_idx, sample_size, replace=False
        )
    else:
        sample = np.tile(alts_idx, numchoosers)

    alts_sample = alternatives.take(sample).copy()

    if alt_index_id:
        # if alt_index_id column name specified, add alt index as a column to interaction dataset
        # permits identification of alternative row in the joined dataset
        alts_sample[alt_index_id] = alts_sample.index

    logger.debug(
        "interaction_dataset pre-merge choosers %s alternatives %s alts_sample %s"
        % (choosers.shape, alternatives.shape, alts_sample.shape)
    )

    # no need to do an expensive merge of alts and choosers
    # we can simply assign repeated chooser values
    for c in choosers.columns:
        c_chooser = (c + "_chooser") if c in alts_sample.columns else c
        alts_sample[c_chooser] = np.repeat(choosers[c].values, sample_size)

    # caller may want this to detect utils that make all alts for a chooser unavailable (e.g. -999)
    if chooser_index_id:
        assert chooser_index_id not in alts_sample
        alts_sample[chooser_index_id] = np.repeat(choosers.index.values, sample_size)

    logger.debug("interaction_dataset merged alts_sample %s" % (alts_sample.shape,))

    return alts_sample


class Nest(object):
    """
    Data for a nest-logit node or leaf

    This object is passed on yield when iterate over nest nodes (branch or leaf)
    The nested logit design is stored in a yaml file as a tree of dict objects,
    but using an object to pass the nest data makes the code a little more readable

    An example nest specification is in the example tour mode choice model
    yaml configuration file - example/configs/tour_mode_choice.yaml.
    """

    def __init__(self, name=None, level=0):
        self.name = name
        self.level = level
        self.product_of_coefficients = 1
        self.ancestors = []
        self.alternatives = None
        self.coefficient = 0

    def print(self):
        print(
            "Nest name: %s level: %s coefficient: %s product_of_coefficients: %s ancestors: %s"
            % (
                self.name,
                self.level,
                self.coefficient,
                self.product_of_coefficients,
                self.ancestors,
            )
        )

    @property
    def is_leaf(self):
        return self.alternatives is None

    @property
    def type(self):
        return "leaf" if self.is_leaf else "node"

    @classmethod
    def nest_types(cls):
        return ["leaf", "node"]


def validate_nest_spec(nest_spec, trace_label):

    keys = []
    duplicates = []
    for nest in each_nest(nest_spec):
        if nest.name in keys:
            logger.error(
                "validate_nest_spec:duplicate nest key '%s' in nest spec - %s"
                % (nest.name, trace_label)
            )
            duplicates.append(nest.name)

        keys.append(nest.name)
        # nest.print()

    if duplicates:
        raise RuntimeError(
            "validate_nest_spec:duplicate nest key/s '%s' in nest spec - %s"
            % (duplicates, trace_label)
        )


def _each_nest(spec, parent_nest, post_order):
    """
    Iterate over each nest or leaf node in the tree (of subtree)

    This internal routine is called by each_nest, which presents a slightly higher level interface

    Parameters
    ----------
    spec : dict
        Nest spec dict tree (or subtree when recursing) from the model spec yaml file
    parent_nest : Nest
        nest of parent node (passed to accumulate level, ancestors, and product_of_coefficients)
    post_order : Bool
        Should we iterate over the nodes of the tree in post-order or pre-order?
        (post-order means we yield the alternatives sub-tree before current node.)

    Yields
    -------
        spec_node : dict
            Nest tree spec dict for this node subtree
        nest : Nest
            Nest object with info about the current node (nest or leaf)
    """
    pre_order = not post_order

    level = parent_nest.level + 1

    if isinstance(spec, dict):
        name = spec["name"]
        coefficient = spec["coefficient"]
        assert isinstance(
            coefficient, (int, float)
        ), "Coefficient '%s' (%s) not a number" % (
            name,
            coefficient,
        )  # forgot to eval coefficient?
        alternatives = [
            a["name"] if isinstance(a, dict) else a for a in spec["alternatives"]
        ]

        nest = Nest(name=name)
        nest.level = parent_nest.level + 1
        nest.coefficient = coefficient
        nest.product_of_coefficients = parent_nest.product_of_coefficients * coefficient
        nest.alternatives = alternatives
        nest.ancestors = parent_nest.ancestors + [name]

        if pre_order:
            yield spec, nest

        # recursively iterate the list of alternatives
        for alternative in spec["alternatives"]:
            for sub_node, sub_nest in _each_nest(alternative, nest, post_order):
                yield sub_node, sub_nest

        if post_order:
            yield spec, nest

    elif isinstance(spec, str):
        name = spec

        nest = Nest(name=name)
        nest.level = parent_nest.level + 1
        nest.product_of_coefficients = parent_nest.product_of_coefficients
        nest.ancestors = parent_nest.ancestors + [name]

        yield spec, nest


def each_nest(nest_spec, type=None, post_order=False):
    """
    Iterate over each nest or leaf node in the tree (of subtree)

    Parameters
    ----------
    nest_spec : dict
        Nest tree dict from the model spec yaml file
    type : str
        Nest class type to yield
        None yields all nests
        'leaf' yields only leaf nodes
        'branch' yields only branch nodes
    post_order : Bool
        Should we iterate over the nodes of the tree in post-order or pre-order?
        (post-order means we yield the alternatives sub-tree before current node.)

    Yields
    -------
        nest : Nest
            Nest object with info about the current node (nest or leaf)
    """
    if type is not None and type not in Nest.nest_types():
        raise RuntimeError("Unknown nest type '%s' in call to each_nest" % type)

    for node, nest in _each_nest(nest_spec, parent_nest=Nest(), post_order=post_order):
        if type is None or (type == nest.type):
            yield nest


def count_nests(nest_spec):
    """
    count the nests in nest_spec, return 0 if nest_spec is none
    """

    def count_each_nest(spec, count):
        if isinstance(spec, dict):
            return (
                count
                + 1
                + sum([count_each_nest(alt, count) for alt in spec["alternatives"]])
            )
        else:
            assert isinstance(spec, str)
            return 1

    return count_each_nest(nest_spec, 0) if nest_spec is not None else 0



# ActivitySim
# See full license in LICENSE.txt.
import logging

from activitysim.core import config, inject, pipeline, simulate, tracing

from .util import estimation

logger = logging.getLogger(__name__)


@inject.step()
def auto_ownership_simulate(households, households_merged, chunk_size, trace_hh_id):
    """
    Auto ownership is a standard model which predicts how many cars a household
    with given characteristics owns
    """
    trace_label = "auto_ownership_simulate"
    model_settings_file_name = "auto_ownership.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    estimator = estimation.manager.begin_estimation("auto_ownership")

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    nest_spec = config.get_logit_model_settings(model_settings)
    constants = config.get_model_constants(model_settings)

    choosers = households_merged.to_frame()

    logger.info("Running %s with %d households", trace_label, len(choosers))

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(choosers)

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = simulate.simple_simulate(
        choosers=choosers,
        spec=model_spec,
        nest_spec=nest_spec,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_label=trace_label,
        trace_choice_name="auto_ownership",
        log_alt_losers=log_alt_losers,
        estimator=estimator,
    )

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, "households", "auto_ownership")
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    households = households.to_frame()

    # no need to reindex as we used all households
    households["auto_ownership"] = choices

    pipeline.replace_table("households", households)

    tracing.print_summary(
        "auto_ownership", households.auto_ownership, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(households, label="auto_ownership", warn_if_empty=True)



import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

parser = argparse.ArgumentParser(description="check activitysim raw_data")
parser.add_argument(
    "raw_data_directory",
    metavar="raw_data_directory",
    type=str,
    nargs=1,
    help=f"path to raw data directory",
)

parser.add_argument(
    "-o", "--output", type=str, metavar="PATH", help="path to output dir"
)

args = parser.parse_args()


input_dir = args.raw_data_directory[0]
output_dir = args.output

print(f"input_dir {input_dir}")
print(f"output_dir {output_dir}")


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def integerize_id_columns(df, table_name):
    columns = ["MAZ", "OMAZ", "DMAZ", "TAZ", "zone_id", "household_id", "HHID"]
    for c in df.columns:
        if c in columns:
            bad = ~(df[c] == df[c].astype(int))
            if bad.any():
                print(f"\n### OOPS ### table {table_name} bad integer column {c}\n")
            df[c] = df[c].astype(int)


def read_csv(file_name, integerize=True):
    df = pd.read_csv(input_path(file_name))

    print(f"read {file_name} {df.shape}")

    return df


def to_csv(df, file_name):
    print(f"writing {file_name} {df.shape} {output_path(file_name)}")
    df.to_csv(output_path(file_name), index=False)


def report_baddies(df, tag, fatal=False):

    if len(df) > 0:
        print(f"\n### OOPS ### {len(df)} {tag}\n")

        # print(f"\n{df}\n")

        if output_dir:
            file_name = f"{tag}.csv"
            print(f"writing {tag} {df.shape} to {output_path(file_name)}")
            df.to_csv(output_path(file_name), index=False)

        if fatal:
            raise RuntimeError(tag)
    else:
        print(f"{len(df)} {tag}")


print(f"input_dir {input_dir} output_dir {output_dir}")

if output_dir and not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


land_use = read_csv("land_use.csv")

# ### check maz.csv against land_use

land_use = land_use.sort_values("MAZ")
maz = read_csv("maz.csv").sort_values("MAZ")

# fatal
missing = land_use.MAZ[~land_use.MAZ.isin(maz.MAZ)]
report_baddies(missing, "land_use_MAZ_not_in_maz_MAZ", fatal=True)

missing = maz.MAZ[~maz.MAZ.isin(land_use.MAZ)]
report_baddies(missing, "maz_MAZ_not_in_land_use_MAZ")

# fatal
missing = land_use.TAZ[~land_use.TAZ.isin(maz.TAZ)]
report_baddies(missing, "land_use_TAZ_not_in_maz_TAZ", fatal=True)

missing = maz.TAZ[~maz.TAZ.isin(land_use.TAZ)]
report_baddies(missing, "maz_TAZ_not_in_land_use_TAZ")

# ### check taz.csv against land_use

land_use = land_use.sort_values("TAZ")
taz = read_csv("taz.csv").sort_values("TAZ")

if output_dir:
    taz.to_csv(output_path("taz.csv"), index=False)

# fatal
missing = land_use.TAZ[~land_use.TAZ.isin(taz.TAZ)]
report_baddies(missing, "land_use_TAZ_not_in_taz_TAZ", fatal=True)

missing = taz.TAZ[~taz.TAZ.isin(land_use.TAZ)]
report_baddies(missing, "taz_TAZ_not_in_land_use_TAZ")

# #########s

#
# maz
#
maz = read_csv("maz.csv").sort_values(["MAZ", "TAZ"])
maz = maz[maz["MAZ"].isin(land_use.MAZ)]
integerize_id_columns(maz, "maz")

assert land_use.MAZ.isin(maz.MAZ).all()
assert land_use.TAZ.isin(maz.TAZ).all()
assert maz.TAZ.isin(land_use.TAZ).all()

#
# taz
#
taz = read_csv("taz.csv").sort_values(["TAZ"])
taz = taz[taz["TAZ"].isin(land_use.TAZ)]
integerize_id_columns(taz, "taz")

assert land_use.TAZ.isin(taz.TAZ).all()

# print(maz.shape)
# print(f"MAZ {len(maz.MAZ.unique())}")
# print(f"TAZ {len(maz.TAZ.unique())}")

#
# households
#
households = read_csv("households.csv")
missing = households[~households["MAZ"].isin(maz.MAZ)]
report_baddies(missing, "household_MAZ_not_in_maz_MAZ")

integerize_id_columns(households, "households")

#
# persons
#
persons = read_csv("persons.csv")
orphans = persons[~persons["household_id"].isin(households.HHID)]
report_baddies(orphans, "persons_not_in_households")

households = households[households["MAZ"].isin(maz.MAZ)]
orphans = persons[~persons["household_id"].isin(households.HHID)]
report_baddies(orphans, "persons_not_in_households_in_maz_MAZ")

integerize_id_columns(persons, "persons")

#
# maz_to_maz_walk and maz_to_maz_bike
#

m2m = read_csv("maz_to_maz_walk.csv")
missing = m2m[~(m2m.OMAZ.isin(maz.MAZ) & m2m.DMAZ.isin(maz.MAZ))]
report_baddies(missing, "maz_to_maz_walk_OMAZ_or_DMAZ_not_in_maz_MAZ")
integerize_id_columns(m2m, "maz_to_maz_walk")

m2m = read_csv("maz_to_maz_bike.csv")
missing = m2m[~(m2m.OMAZ.isin(maz.MAZ) & m2m.DMAZ.isin(maz.MAZ))]
report_baddies(missing, "maz_to_maz_bike_OMAZ_or_DMAZ_not_in_maz_MAZ")
integerize_id_columns(m2m, "maz_to_maz_bike")


#
# skims
#
omx_infile_name = "skims.omx"
skim_data_type = np.float32

omx_in = omx.open_file(input_path(omx_infile_name), "r")
print(f"omx_in shape {omx_in.shape()}")

print(f"{len(omx_in.listMappings())} mappings in skims")

for m in omx_in.listMappings():
    print(f"found mapping '{m}' in skims")
assert len(omx_in.listMappings()) == 0


# assert omx_in.shape() == (len(taz), len(taz))

omx_in.close()



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing
from activitysim.core.util import assign_in_place, reindex

from .util import estimation
from .util.vectorize_tour_scheduling import vectorize_joint_tour_scheduling

logger = logging.getLogger(__name__)


@inject.step()
def joint_tour_scheduling(tours, persons_merged, tdd_alts, chunk_size, trace_hh_id):
    """
    This model predicts the departure time and duration of each joint tour
    """
    trace_label = "joint_tour_scheduling"

    model_settings_file_name = "joint_tour_scheduling.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    tours = tours.to_frame()
    joint_tours = tours[tours.tour_category == "joint"]

    # - if no joint tours
    if joint_tours.shape[0] == 0:
        tracing.no_results(trace_label)
        return

    # use inject.get_table as this won't exist if there are no joint_tours
    joint_tour_participants = inject.get_table("joint_tour_participants").to_frame()

    persons_merged = persons_merged.to_frame()

    logger.info("Running %s with %d joint tours", trace_label, joint_tours.shape[0])

    # it may seem peculiar that we are concerned with persons rather than households
    # but every joint tour is (somewhat arbitrarily) assigned a "primary person"
    # some of whose characteristics are used in the spec
    # and we get household attributes along with person attributes in persons_merged
    persons_merged = persons_merged[persons_merged.num_hh_joint_tours > 0]

    # since a households joint tours each potentially different participants
    # they may also have different joint tour masks (free time of all participants)
    # so we have to either chunk processing by joint_tour_num and build timetable by household
    # or build timetables by unique joint_tour

    constants = config.get_model_constants(model_settings)

    # - run preprocessor to annotate choosers
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:

        locals_d = {}
        if constants is not None:
            locals_d.update(constants)

        expressions.assign_columns(
            df=joint_tours,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label,
        )

    timetable = inject.get_injectable("timetable")

    estimator = estimation.manager.begin_estimation("joint_tour_scheduling")

    model_spec = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df, model_settings)
        timetable.begin_transaction(estimator)

    choices = vectorize_joint_tour_scheduling(
        joint_tours,
        joint_tour_participants,
        persons_merged,
        tdd_alts,
        timetable,
        spec=model_spec,
        model_settings=model_settings,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_label=trace_label,
    )

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, "tours", "tdd")
        estimator.write_override_choices(choices)
        estimator.end_estimation()

        # update timetable to reflect the override choices (assign tours in tour_num order)
        timetable.rollback()
        for tour_num, nth_tours in joint_tours.groupby("tour_num", sort=True):
            nth_participants = joint_tour_participants[
                joint_tour_participants.tour_id.isin(nth_tours.index)
            ]

            estimator.log(
                "assign timetable for %s participants in %s tours with tour_num %s"
                % (len(nth_participants), len(nth_tours), tour_num)
            )
            # - update timetables of all joint tour participants
            timetable.assign(
                nth_participants.person_id, reindex(choices, nth_participants.tour_id)
            )

    timetable.replace_table()

    # choices are tdd alternative ids
    # we want to add start, end, and duration columns to tours, which we have in tdd_alts table
    choices = pd.merge(
        choices.to_frame("tdd"), tdd_alts, left_on=["tdd"], right_index=True, how="left"
    )

    assign_in_place(tours, choices)
    pipeline.replace_table("tours", tours)

    # updated df for tracing
    joint_tours = tours[tours.tour_category == "joint"]

    if trace_hh_id:
        tracing.trace_df(
            joint_tours, label="joint_tour_scheduling", slicer="household_id"
        )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, simulate, tracing

from . import estimation
from . import vectorize_tour_scheduling as vts

logger = logging.getLogger(__name__)


def run_tour_scheduling(
    model_name,
    chooser_tours,
    persons_merged,
    tdd_alts,
    tour_segment_col,
    chunk_size,
    trace_hh_id,
):

    trace_label = model_name
    model_settings_file_name = f"{model_name}.yaml"

    model_settings = config.read_model_settings(model_settings_file_name)

    if "LOGSUM_SETTINGS" in model_settings:
        logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])
        logsum_columns = logsum_settings.get("LOGSUM_CHOOSER_COLUMNS", [])
    else:
        logsum_columns = []

    # - filter chooser columns for both logsums and simulate
    model_columns = model_settings.get("SIMULATE_CHOOSER_COLUMNS", [])
    chooser_columns = logsum_columns + [
        c for c in model_columns if c not in logsum_columns
    ]

    persons_merged = expressions.filter_chooser_columns(persons_merged, chooser_columns)

    timetable = inject.get_injectable("timetable")

    # - run preprocessor to annotate choosers
    preprocessor_settings = model_settings.get("preprocessor", None)
    if preprocessor_settings:
        locals_d = {"tt": timetable}
        locals_d.update(config.get_model_constants(model_settings))

        expressions.assign_columns(
            df=chooser_tours,
            model_settings=preprocessor_settings,
            locals_dict=locals_d,
            trace_label=trace_label,
        )

    estimators = {}
    if "TOUR_SPEC_SEGMENTS" in model_settings:
        # load segmented specs
        spec_segment_settings = model_settings.get("SPEC_SEGMENTS", {})
        specs = {}
        for spec_segment_name, spec_settings in spec_segment_settings.items():

            bundle_name = f"{model_name}_{spec_segment_name}"

            # estimator for this tour_segment
            estimator = estimation.manager.begin_estimation(
                model_name=bundle_name, bundle_name=bundle_name
            )

            spec_file_name = spec_settings["SPEC"]
            model_spec = simulate.read_model_spec(file_name=spec_file_name)
            coefficients_df = simulate.read_model_coefficients(spec_settings)
            specs[spec_segment_name] = simulate.eval_coefficients(
                model_spec, coefficients_df, estimator
            )

            if estimator:
                estimators[spec_segment_name] = estimator  # add to local list
                estimator.write_model_settings(model_settings, model_settings_file_name)
                estimator.write_spec(spec_settings)
                estimator.write_coefficients(coefficients_df, spec_settings)

        # - spec dict segmented by primary_purpose
        tour_segment_settings = model_settings.get("TOUR_SPEC_SEGMENTS", {})
        tour_segments = {}
        for tour_segment_name, spec_segment_name in tour_segment_settings.items():
            tour_segments[tour_segment_name] = {}
            tour_segments[tour_segment_name]["spec_segment_name"] = spec_segment_name
            tour_segments[tour_segment_name]["spec"] = specs[spec_segment_name]
            tour_segments[tour_segment_name]["estimator"] = estimators.get(
                spec_segment_name
            )

        # default tour_segment_col to 'tour_type' if segmented spec and tour_segment_col not specified
        if tour_segment_col is None and tour_segments:
            tour_segment_col = "tour_type"

    else:
        # unsegmented spec
        assert "SPEC_SEGMENTS" not in model_settings
        assert "TOUR_SPEC_SEGMENTS" not in model_settings
        assert tour_segment_col is None

        estimator = estimation.manager.begin_estimation(model_name)

        spec_file_name = model_settings["SPEC"]
        model_spec = simulate.read_model_spec(file_name=spec_file_name)
        coefficients_df = simulate.read_model_coefficients(model_settings)
        model_spec = simulate.eval_coefficients(model_spec, coefficients_df, estimator)

        if estimator:
            estimators[None] = estimator  # add to local list
            estimator.write_model_settings(model_settings, model_settings_file_name)
            estimator.write_spec(model_settings)
            estimator.write_coefficients(coefficients_df, model_settings)

        # - non_mandatory tour scheduling is not segmented by tour type
        tour_segments = {"spec": model_spec, "estimator": estimator}

    if estimators:
        timetable.begin_transaction(list(estimators.values()))

    logger.info(f"Running {model_name} with %d tours", len(chooser_tours))
    choices = vts.vectorize_tour_scheduling(
        chooser_tours,
        persons_merged,
        tdd_alts,
        timetable,
        tour_segments=tour_segments,
        tour_segment_col=tour_segment_col,
        model_settings=model_settings,
        chunk_size=chunk_size,
        trace_label=trace_label,
    )

    if estimators:
        # overrride choices for all estimators
        choices_list = []
        for spec_segment_name, estimator in estimators.items():
            if spec_segment_name:
                model_choices = choices[(chooser_tours.tour_type == spec_segment_name)]
            else:
                model_choices = choices

            estimator.write_choices(model_choices)
            override_choices = estimator.get_survey_values(
                model_choices, "tours", "tdd"
            )
            estimator.write_override_choices(override_choices)

            choices_list.append(override_choices)
            estimator.end_estimation()
        choices = pd.concat(choices_list)

        # update timetable to reflect the override choices (assign tours in tour_num order)
        timetable.rollback()
        for tour_num, nth_tours in chooser_tours.groupby("tour_num", sort=True):
            timetable.assign(
                window_row_ids=nth_tours["person_id"],
                tdds=choices.reindex(nth_tours.index),
            )

    timetable.replace_table()

    # choices are tdd alternative ids
    # we want to add start, end, and duration columns to tours, which we have in tdd_alts table
    choices = pd.merge(
        choices.to_frame("tdd"), tdd_alts, left_on=["tdd"], right_index=True, how="left"
    )

    return choices



# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import range

import pandas as pd

from activitysim.core import inject, mem, pipeline, tracing
from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


@inject.table()
def households(households_sample_size, override_hh_ids, trace_hh_id):

    df_full = read_input_table("households")
    tot_households = df_full.shape[0]

    logger.info("full household list contains %s households" % tot_households)

    households_sliced = False

    # only using households listed in override_hh_ids
    if override_hh_ids is not None:

        # trace_hh_id will not used if it is not in list of override_hh_ids
        logger.info(
            "override household list containing %s households" % len(override_hh_ids)
        )

        df = df_full[df_full.index.isin(override_hh_ids)]
        households_sliced = True

        if df.shape[0] < len(override_hh_ids):
            logger.info(
                "found %s of %s households in override household list"
                % (df.shape[0], len(override_hh_ids))
            )

        if df.shape[0] == 0:
            raise RuntimeError("No override households found in store")

    # if we are tracing hh exclusively
    elif trace_hh_id and households_sample_size == 1:

        # df contains only trace_hh (or empty if not in full store)
        df = tracing.slice_ids(df_full, trace_hh_id)
        households_sliced = True

    # if we need a subset of full store
    elif tot_households > households_sample_size > 0:

        logger.info(
            "sampling %s of %s households" % (households_sample_size, tot_households)
        )

        """
        Because random seed is set differently for each step, sampling of households using
        Random.global_rng would sample differently depending upon which step it was called from.
        We use a one-off rng seeded with the pseudo step name 'sample_households' to provide
        repeatable sampling no matter when the table is loaded.

        Note that the external_rng is also seeded with base_seed so the sample will (rightly) change
        if the pipeline rng's base_seed is changed
        """

        prng = pipeline.get_rn_generator().get_external_rng("sample_households")
        df = df_full.take(
            prng.choice(len(df_full), size=households_sample_size, replace=False)
        )
        households_sliced = True

        # if tracing and we missed trace_hh in sample, but it is in full store
        if trace_hh_id and trace_hh_id not in df.index and trace_hh_id in df_full.index:
            # replace first hh in sample with trace_hh
            logger.debug(
                "replacing household %s with %s in household sample"
                % (df.index[0], trace_hh_id)
            )
            df_hh = df_full.loc[[trace_hh_id]]
            df = pd.concat([df_hh, df[1:]])

    else:
        df = df_full

    # persons table
    inject.add_injectable("households_sliced", households_sliced)

    if "sample_rate" not in df.columns:
        if households_sample_size == 0:
            sample_rate = 1
        else:
            sample_rate = round(households_sample_size / tot_households, 3)

        df["sample_rate"] = sample_rate

    logger.info("loaded households %s" % (df.shape,))

    # replace table function with dataframe
    inject.add_table("households", df)

    pipeline.get_rn_generator().add_channel("households", df)

    tracing.register_traceable_table("households", df)
    if trace_hh_id:
        tracing.trace_df(df, "raw.households", warn_if_empty=True)

    return df


# this is a common merge so might as well define it once here and use it
@inject.table()
def households_merged(households, land_use, accessibility):
    return inject.merge_tables(
        households.name, tables=[households, land_use, accessibility]
    )


inject.broadcast("households", "persons", cast_index=True, onto_on="household_id")

# this would be accessibility around the household location - be careful with
# this one as accessibility at some other location can also matter
inject.broadcast("accessibility", "households", cast_index=True, onto_on="home_zone_id")



# ActivitySim
# See full license in LICENSE.txt.

import os.path

import pandas as pd
import pandas.testing as pdt
import pytest
import yaml

from activitysim.core import chunk, config, inject, simulate

from .. import cdap


@pytest.fixture(scope="module")
def data_dir():
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture(scope="module")
def people(data_dir):
    return pd.read_csv(os.path.join(data_dir, "people.csv"), index_col="id")


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


@pytest.fixture(scope="module")
def model_settings(configs_dir):
    yml_file = os.path.join(configs_dir, "cdap.yaml")
    with open(yml_file) as f:
        model_settings = yaml.load(f, Loader=yaml.loader.SafeLoader)
    return model_settings


@pytest.fixture(scope="module")
def configs_dir():
    return os.path.join(os.path.dirname(__file__), "configs")


def setup_function():
    configs_dir = os.path.join(os.path.dirname(__file__), "configs")
    inject.add_injectable("configs_dir", configs_dir)
    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)


def test_bad_coefficients():

    coefficients = pd.read_csv(
        config.config_file_path("cdap_interaction_coefficients.csv"), comment="#"
    )
    coefficients = cdap.preprocess_interaction_coefficients(coefficients)

    coefficients.loc[2, "activity"] = "AA"

    with pytest.raises(RuntimeError) as excinfo:
        coefficients = cdap.preprocess_interaction_coefficients(coefficients)
    assert "Expect only M, N, or H" in str(excinfo.value)


def test_assign_cdap_rank(people, model_settings):

    person_type_map = model_settings.get("PERSON_TYPE_MAP", {})

    with chunk.chunk_log("test_assign_cdap_rank", base=True):
        cdap.assign_cdap_rank(people, person_type_map)

    expected = pd.Series(
        [1, 1, 1, 2, 2, 1, 3, 1, 2, 1, 3, 2, 1, 3, 2, 4, 1, 3, 4, 2], index=people.index
    )

    pdt.assert_series_equal(
        people["cdap_rank"], expected, check_dtype=False, check_names=False
    )


def test_individual_utilities(people, model_settings):

    cdap_indiv_and_hhsize1 = simulate.read_model_spec(
        file_name="cdap_indiv_and_hhsize1.csv"
    )

    person_type_map = model_settings.get("PERSON_TYPE_MAP", {})

    with chunk.chunk_log("test_individual_utilities", base=True):
        cdap.assign_cdap_rank(people, person_type_map)
        individual_utils = cdap.individual_utilities(
            people, cdap_indiv_and_hhsize1, locals_d=None
        )

    individual_utils = individual_utils[["M", "N", "H"]]

    expected = pd.DataFrame(
        [
            [2, 0, 0],  # person 1
            [0, 0, 1],  # person 2
            [3, 0, 0],  # person 3
            [3, 0, 0],  # person 4
            [0, 1, 0],  # person 5
            [1, 0, 0],  # person 6
            [1, 0, 0],  # person 7
            [0, 2, 0],  # person 8
            [0, 0, 1],  # person 9
            [2, 0, 0],  # person 10
            [0, 0, 3],  # person 11
            [0, 0, 2],  # person 12
            [3, 0, 0],  # person 13
            [1, 0, 0],  # person 14
            [0, 4, 0],  # person 15
            [0, 4, 0],  # person 16
            [0, 0, 4],  # person 17
            [0, 0, 5],  # person 18
            [50, 0, 4],  # person 19
            [2, 0, 0],  # person 20
        ],
        index=people.index,
        columns=cdap_indiv_and_hhsize1.columns,
    )

    pdt.assert_frame_equal(
        individual_utils, expected, check_dtype=False, check_names=False
    )


def test_build_cdap_spec_hhsize2(people, model_settings):

    hhsize = 2
    cdap_indiv_and_hhsize1 = simulate.read_model_spec(
        file_name="cdap_indiv_and_hhsize1.csv"
    )

    interaction_coefficients = pd.read_csv(
        config.config_file_path("cdap_interaction_coefficients.csv"), comment="#"
    )
    interaction_coefficients = cdap.preprocess_interaction_coefficients(
        interaction_coefficients
    )

    person_type_map = model_settings.get("PERSON_TYPE_MAP", {})

    with chunk.chunk_log("test_build_cdap_spec_hhsize2", base=True):
        cdap.assign_cdap_rank(people, person_type_map)
        indiv_utils = cdap.individual_utilities(
            people, cdap_indiv_and_hhsize1, locals_d=None
        )

        choosers = cdap.hh_choosers(indiv_utils, hhsize=hhsize)

        spec = cdap.build_cdap_spec(
            interaction_coefficients, hhsize=hhsize, cache=False
        )

        # pandas.dot depends on column names of expression_values matching spec index values
        # expressions should have been uniquified when spec was read
        assert spec.index.is_unique

        vars = simulate.eval_variables(spec.index, choosers)
        assert (spec.index.values == vars.columns.values).all()

    # spec = spec.astype(np.float64)

    utils = vars.dot(spec)

    expected = pd.DataFrame(
        [
            [0, 3, 0, 3, 7, 3, 0, 3, 0],  # household 3
            [0, 0, 1, 1, 1, 2, 0, 0, 2],  # household 4
        ],
        index=[3, 4],
        columns=["HH", "HM", "HN", "MH", "MM", "MN", "NH", "NM", "NN"],
    ).astype("float")

    pdt.assert_frame_equal(utils, expected, check_names=False)



# ActivitySim
# See full license in LICENSE.txt.
# from builtins import int

import logging
import multiprocessing
import os
from abc import ABC, abstractmethod

import numpy as np
import openmatrix as omx

from activitysim.core import config, inject, skim_dictionary, tracing, util

logger = logging.getLogger(__name__)


class SkimData(object):
    """
    A facade for 3D skim data exposing numpy indexing and shape
    The primary purpose is to document and police the api used to access skim data
    Subclasses using a different backing store to perform additional/alternative
    only need to implement the methods exposed here.

    For instance, to open/close memmapped files just in time, or to access backing data via an alternate api
    """

    def __init__(self, skim_data):
        """
        skim_data is an np.ndarray or anything that implements the methods/properties of this class

        Parameters
        ----------
        skim_data : np.ndarray or quack-alike
        """
        self._skim_data = skim_data

    def __getitem__(self, indexes):
        if len(indexes) != 3:
            raise ValueError(f"number of indexes ({len(indexes)}) should be 3")
        return self._skim_data[indexes]

    @property
    def shape(self):
        """
        Returns
        -------
        list-like shape tuple as returned by numpy.shape
        """
        return self._skim_data.shape


class SkimInfo(object):
    def __init__(self, skim_tag, network_los):
        """

        skim_tag:           str             (e.g. 'TAZ')
        dtype_name:         str             (e.g. 'float32')
        omx_manifest:       dict            dict mapping { omx_key: omx_file_name }
        omx_shape:          2D tuple        shape of omx matrix: (<number_of_zones>, <number_of_zones>)
        num_skims:          int             total number of individual skim matrices in omx files
        skim_data_shape:    3D tuple        (num_skims, omx_shape[0], omx_shape[1]) if ROW_MAJOR_LAYOUT
        offset_map:         dict or None    1D ndarray as returned by omx_file.mapentries, if omx file has mappings
        offset_map_name:    str             name of offset_map in omx_filecorresponding to offset_map, if there was one
        omx_keys:       dict            dict mapping skim key (str or tuple) to skim key in omx file
                                            {DISTWALK: DISTWALK,
                                            ('DRV_COM_WLK_BOARDS', 'AM'): DRV_COM_WLK_BOARDS__AM, ...}
        base_keys:          list of str     e.g. 'BIKEDIST' or 'SOVTOLL_VTOLL' (base key of 3d skim)
        block_offsets:      dict            dict mapping skim key tuple to offset

        Parameters
        ----------
        skim_tag
        """

        self.network_los = network_los
        self.skim_tag = skim_tag
        self.dtype_name = network_los.skim_dtype_name

        self.omx_manifest = None
        self.omx_shape = None
        self.num_skims = None
        self.skim_data_shape = None
        self.offset_map_name = None
        self.offset_map = None
        self.omx_keys = None
        self.base_keys = None
        self.block_offsets = None

        self.load_skim_info(skim_tag)

    def load_skim_info(self, skim_tag):
        """
        Read omx files for skim <skim_tag> (e.g. 'TAZ') and build skim_info dict

        Parameters
        ----------
        skim_tag: str

        """

        omx_file_names = self.network_los.omx_file_names(skim_tag)

        self.omx_file_paths = config.expand_input_file_list(omx_file_names)

        # ignore any 3D skims not in skim_time_periods
        # specifically, load all skims except those with key2 not in dim3_tags_to_load
        skim_time_periods = self.network_los.skim_time_periods
        dim3_tags_to_load = skim_time_periods and skim_time_periods["labels"]

        self.omx_manifest = {}  # dict mapping { omx_key: skim_name }

        for omx_file_path in self.omx_file_paths:

            logger.debug(f"load_skim_info {skim_tag} reading {omx_file_path}")

            with omx.open_file(omx_file_path) as omx_file:

                # fixme call to omx_file.shape() failing in windows p3.5
                if self.omx_shape is None:
                    self.omx_shape = tuple(
                        int(i) for i in omx_file.shape()
                    )  # sometimes omx shape are floats!
                else:
                    assert self.omx_shape == tuple(
                        int(i) for i in omx_file.shape()
                    ), f"Mismatch shape {self.omx_shape} != {omx_file.shape()}"

                for skim_name in omx_file.listMatrices():
                    assert (
                        skim_name not in self.omx_manifest
                    ), f"duplicate skim '{skim_name}' found in {self.omx_manifest[skim_name]} and {omx_file}"
                    self.omx_manifest[skim_name] = omx_file_path

                for m in omx_file.listMappings():
                    if self.offset_map is None:
                        self.offset_map_name = m
                        self.offset_map = omx_file.mapentries(self.offset_map_name)
                        assert len(self.offset_map) == self.omx_shape[0]
                    else:
                        # don't really expect more than one, but ok if they are all the same
                        if not (self.offset_map == omx_file.mapentries(m)):
                            raise RuntimeError(
                                f"Multiple mappings in omx file: {self.offset_map_name} != {m}"
                            )

        # - omx_keys dict maps skim key to omx_key
        # DISTWALK: DISTWALK
        # ('DRV_COM_WLK_BOARDS', 'AM'): DRV_COM_WLK_BOARDS__AM, ...
        self.omx_keys = dict()
        for skim_name in self.omx_manifest.keys():
            key1, sep, key2 = skim_name.partition("__")

            # - ignore composite tags not in dim3_tags_to_load
            if dim3_tags_to_load and sep and key2 not in dim3_tags_to_load:
                continue

            skim_key = (key1, key2) if sep else key1

            self.omx_keys[skim_key] = skim_name

        self.num_skims = len(self.omx_keys)

        # - key1_subkeys dict maps key1 to dict of subkeys with that key1
        # DIST: {'DIST': 0}
        # DRV_COM_WLK_BOARDS: {'MD': 1, 'AM': 0, 'PM': 2}, ...
        key1_subkeys = dict()
        for skim_key, omx_key in self.omx_keys.items():
            if isinstance(skim_key, tuple):
                key1, key2 = skim_key
            else:
                key1 = key2 = skim_key
            key2_dict = key1_subkeys.setdefault(key1, {})
            key2_dict[key2] = len(key2_dict)

        key1_block_offsets = dict()
        offset = 0
        for key1, v in key1_subkeys.items():
            num_subkeys = len(v)
            key1_block_offsets[key1] = offset
            offset += num_subkeys

        # - block_offsets dict maps skim_key to offset of omx matrix
        # DIST: 0,
        # ('DRV_COM_WLK_BOARDS', 'AM'): 3,
        # ('DRV_COM_WLK_BOARDS', 'MD') 4, ...
        self.block_offsets = dict()
        for skim_key in self.omx_keys:

            if isinstance(skim_key, tuple):
                key1, key2 = skim_key
            else:
                key1 = key2 = skim_key

            key1_offset = key1_block_offsets[key1]
            key2_relative_offset = key1_subkeys.get(key1).get(key2)
            self.block_offsets[skim_key] = key1_offset + key2_relative_offset

        if skim_dictionary.ROW_MAJOR_LAYOUT:
            self.skim_data_shape = (
                self.num_skims,
                self.omx_shape[0],
                self.omx_shape[1],
            )
        else:
            self.skim_data_shape = self.omx_shape + (self.num_skims,)

        # list of base keys (keys
        self.base_keys = tuple(k for k in key1_block_offsets.keys())

    def print(self):
        print(f"SkimInfo for {self.skim_tag}")
        print(f"omx_shape {self.omx_shape}")
        print(f"num_skims {self.num_skims}")
        print(f"skim_data_shape {self.skim_data_shape}")
        print(f"offset_map_name {self.offset_map_name}")
        # print(f"omx_manifest {self.omx_manifest}")
        # print(f"offset_map {self.offset_map}")
        # print(f"omx_keys {self.omx_keys}")
        # print(f"base_keys {self.base_keys}")
        # print(f"block_offsets {self.block_offsets}")


class AbstractSkimFactory(ABC):
    """
    Provide access to skim data from store.

    load_skim_info(skim_tag: str): dict
        Read omx files for skim <skim_tag> (e.g. 'TAZ') and build skim_info dict

    get_skim_data(skim_tag: str, skim_info: dict): SkimData
        Read skim data from backing store and return it as a 3D ndarray quack-alike SkimData object

    allocate_skim_buffer(skim_info, shared: bool): 1D array buffer sized for 3D SkimData
        Allocate a ram skim buffer (ndarray or multiprocessing.Array) to use as frombuffer for SkimData

    """

    def __init__(self, network_los):
        self.network_los = network_los

    @property
    def supports_shared_data_for_multiprocessing(self):
        """
        Does subclass support shareable data for multiprocessing

        Returns
        -------
        boolean
        """
        return False

    def allocate_skim_buffer(self, skim_info, shared=False):
        """
        For multiprocessing
        """
        assert False, "Not supported"

    def _skim_data_from_buffer(self, skim_info, skim_buffer):
        assert False, "Not supported"

    def _memmap_skim_data_path(self, skim_tag):
        return os.path.join(config.get_cache_dir(), f"cached_{skim_tag}.mmap")

    def load_skim_info(self, skim_tag):
        return SkimInfo(skim_tag, self.network_los)

    def _read_skims_from_omx(self, skim_info, skim_data):
        """
        read skims from omx file into skim_data
        """

        skim_tag = skim_info.skim_tag
        omx_keys = skim_info.omx_keys
        omx_manifest = skim_info.omx_manifest  # dict mapping { omx_key: skim_name }

        for omx_file_path in skim_info.omx_file_paths:

            num_skims_loaded = 0

            logger.info(f"_read_skims_from_omx {omx_file_path}")

            # read skims into skim_data
            with omx.open_file(omx_file_path) as omx_file:
                for skim_key, omx_key in omx_keys.items():

                    if omx_manifest[omx_key] == omx_file_path:

                        offset = skim_info.block_offsets[skim_key]
                        logger.debug(
                            f"_read_skims_from_omx file {omx_file_path} omx_key {omx_key} "
                            f"skim_key {skim_key} to offset {offset}"
                        )

                        if skim_dictionary.ROW_MAJOR_LAYOUT:
                            a = skim_data[offset, :, :]
                        else:
                            a = skim_data[:, :, offset]

                        # this will trigger omx readslice to read and copy data to skim_data's buffer
                        omx_data = omx_file[omx_key]
                        a[:] = omx_data[:]

                        num_skims_loaded += 1

            logger.info(
                f"_read_skims_from_omx loaded {num_skims_loaded} skims from {omx_file_path}"
            )

    def _open_existing_readonly_memmap_skim_cache(self, skim_info):
        """
        read cached memmapped skim data from canonically named cache file(s) in output directory into skim_data
        return True if it was there and we read it, return False if not found
        """

        dtype = np.dtype(skim_info.dtype_name)

        skim_cache_path = self._memmap_skim_data_path(skim_info.skim_tag)

        if not os.path.isfile(skim_cache_path):
            logger.warning(f"read_skim_cache file not found: {skim_cache_path}")
            return None

        logger.info(
            f"reading skim cache {skim_info.skim_tag} {skim_info.skim_data_shape} from {skim_cache_path}"
        )

        try:
            data = np.memmap(
                skim_cache_path, shape=skim_info.skim_data_shape, dtype=dtype, mode="r"
            )
        except Exception as e:
            logger.warning(
                f"{type(e).__name__} reading {skim_info.skim_tag} skim_cache {skim_cache_path}:  {str(e)}"
            )
            logger.warning(
                f"ignoring incompatible {skim_info.skim_tag} skim_cache {skim_cache_path}"
            )
            return None

        return data

    def _create_empty_writable_memmap_skim_cache(self, skim_info):
        """
        write skim data from skim_data to canonically named cache file(s) in output directory
        """

        dtype = np.dtype(skim_info.dtype_name)

        skim_cache_path = self._memmap_skim_data_path(skim_info.skim_tag)

        logger.info(
            f"writing skim cache {skim_info.skim_tag} {skim_info.skim_data_shape} to {skim_cache_path}"
        )

        data = np.memmap(
            skim_cache_path, shape=skim_info.skim_data_shape, dtype=dtype, mode="w+"
        )

        return data

    def copy_omx_to_mmap_file(self, skim_info):

        skim_data = self._create_empty_writable_memmap_skim_cache(skim_info)
        self._read_skims_from_omx(skim_info, skim_data)
        skim_data._mmap.close()
        del skim_data


class NumpyArraySkimFactory(AbstractSkimFactory):
    def __init__(self, network_los):
        super().__init__(network_los)

    @property
    def supports_shared_data_for_multiprocessing(self):
        return True

    def allocate_skim_buffer(self, skim_info, shared=False):
        """
        Allocate a ram skim buffer to use as frombuffer for SkimData
        If shared is True, return a shareable multiprocessing.RawArray, otherwise a numpy.ndarray

        Parameters
        ----------
        skim_info: dict
        shared: boolean

        Returns
        -------
        multiprocessing.RawArray or numpy.ndarray
        """

        assert (
            shared == self.network_los.multiprocess()
        ), f"NumpyArraySkimFactory.allocate_skim_buffer shared {shared} multiprocess {not shared}"

        dtype_name = skim_info.dtype_name
        dtype = np.dtype(dtype_name)

        # multiprocessing.RawArray argument buffer_size must be int, not np.int64
        buffer_size = util.iprod(skim_info.skim_data_shape)

        csz = buffer_size * dtype.itemsize
        logger.info(
            f"allocate_skim_buffer shared {shared} {skim_info.skim_tag} shape {skim_info.skim_data_shape} "
            f"total size: {util.INT(csz)} ({util.GB(csz)})"
        )

        if shared:
            if dtype_name == "float64":
                typecode = "d"
            elif dtype_name == "float32":
                typecode = "f"
            else:
                raise RuntimeError(
                    "allocate_skim_buffer unrecognized dtype %s" % dtype_name
                )

            buffer = multiprocessing.RawArray(typecode, buffer_size)
        else:
            buffer = np.zeros(buffer_size, dtype=dtype)

        return buffer

    def _skim_data_from_buffer(self, skim_info, skim_buffer):
        """
        return a numpy ndarray using skim_buffer as backing store

        Parameters
        ----------
        skim_info
        skim_buffer

        Returns
        -------

        """

        dtype = np.dtype(skim_info.dtype_name)
        assert len(skim_buffer) == util.iprod(skim_info.skim_data_shape)
        skim_data = np.frombuffer(skim_buffer, dtype=dtype).reshape(
            skim_info.skim_data_shape
        )
        return skim_data

    def load_skims_to_buffer(self, skim_info, skim_buffer):
        """
        Load skims from disk store (omx or cache) into ram skim buffer (multiprocessing.RawArray or numpy.ndarray)

        Parameters
        ----------
        skim_info: doct
        skim_buffer: 1D buffer sized to hold all skims (multiprocessing.RawArray or numpy.ndarray)
        """

        read_cache = self.network_los.setting("read_skim_cache", False)
        write_cache = self.network_los.setting("write_skim_cache", False)

        skim_data = self._skim_data_from_buffer(skim_info, skim_buffer)
        assert skim_data.shape == skim_info.skim_data_shape

        if read_cache:
            # returns None if cache file not found
            cache_data = self._open_existing_readonly_memmap_skim_cache(skim_info)

            # copy memmapped cache to RAM numpy ndarray
            if cache_data is not None:
                assert cache_data.shape == skim_data.shape
                np.copyto(skim_data, cache_data)
                cache_data._mmap.close()
                del cache_data
                return

        # read omx skims into skim_buffer (np array)
        self._read_skims_from_omx(skim_info, skim_data)

        if write_cache:
            cache_data = self._create_empty_writable_memmap_skim_cache(skim_info)
            np.copyto(cache_data, skim_data)
            cache_data._mmap.close()
            del cache_data

            # bug - do we need to close it?

        logger.info(
            f"load_skims_to_buffer {skim_info.skim_tag} shape {skim_data.shape}"
        )

    def get_skim_data(self, skim_tag, skim_info):
        """
        Read skim data from backing store and return it as a 3D ndarray quack-alike SkimData object

        Parameters
        ----------
        skim_tag: str
        skim_info: string

        Returns
        -------
        SkimData
        """

        data_buffers = inject.get_injectable("data_buffers", None)
        if data_buffers:
            # we assume any existing skim buffers will already have skim data loaded into them
            logger.info(
                f"get_skim_data {skim_tag} using existing shared skim_buffers for skims"
            )
            skim_buffer = data_buffers[skim_tag]
        else:
            skim_buffer = self.allocate_skim_buffer(skim_info, shared=False)
            self.load_skims_to_buffer(skim_info, skim_buffer)

        skim_data = SkimData(self._skim_data_from_buffer(skim_info, skim_buffer))

        logger.info(
            f"get_skim_data {skim_tag} {type(skim_data).__name__} shape {skim_data.shape}"
        )

        return skim_data


class JitMemMapSkimData(SkimData):
    """
    SkimData subclass for just-in-time memmap.

    Since opening a memmap is fast, open the memmap read the data on demand and immediately close it.
    This essentially eliminates RAM usage, but it means we are loading the data every time we access the skim,
    which may be significantly slower, depending on patterns of usage.
    """

    def __init__(self, skim_cache_path, skim_info):
        super().__init__(skim_info)
        self.skim_cache_path = skim_cache_path
        self.dtype = np.dtype(skim_info.dtype_name)
        self._shape = skim_info.skim_data_shape

    def __getitem__(self, indexes):
        assert len(indexes) == 3, f"number of indexes ({len(indexes)}) should be 3"
        # open memmap
        data = np.memmap(
            self.skim_cache_path, shape=self._shape, dtype=self.dtype, mode="r"
        )
        # dereference skim values
        result = data[indexes]
        # closing memmap's underlying mmap frees data read into (not really needed as we are exiting scope)
        data._mmap.close()
        return result

    @property
    def shape(self):
        return self._shape


class MemMapSkimFactory(AbstractSkimFactory):
    """
    The numpy.memmap docs states: The memmap object can be used anywhere an ndarray is accepted.
    You might think that since memmap duck-types ndarray, we could simply wrap it in a SkimData object.

    But, as the numpy.memmap docs also say: "Memory-mapped files are used for accessing
    small segments of large files on disk, without reading the entire file into memory."

    The words "small segments" are not accidental, because, as you gradually access all the parts
    of the memmapped array, memory usage increases as all the memory is loaded into RAM.

    Under this scenario, the MemMapSkimFactory operates as a just-in-time loader, with no net savings
    in RAM footprint (other than potentially avoiding loading any unused skims).

    Alternatively, since opening a memmap is fast, you could just open the memmap read the data on demand,
    and immediately close it. This essentially eliminates RAM usage, but it means you are loading the data
    every time you access the skim, which, depending on you patterns of usage, may or may not be acceptable.

    """

    def __init__(self, network_los):
        super().__init__(network_los)

    def get_skim_data(self, skim_tag, skim_info):
        """
        Read skim data from backing store and return it as a 3D ndarray quack-alike SkimData object
        (either a JitMemMapSkimData or a memmap backed SkimData object)

        Parameters
        ----------
        skim_tag: str
        skim_info: string

        Returns
        -------
        SkimData or subclass
        """

        # don't expect legacy shared memory buffers
        assert not inject.get_injectable("data_buffers", {}).get(skim_tag)

        skim_cache_path = self._memmap_skim_data_path(skim_tag)
        if not os.path.isfile(skim_cache_path):
            self.copy_omx_to_mmap_file(skim_info)

        JIT = True  # FIXME - this should be a network_los setting, along with selection of the factory?
        if JIT:
            skim_data = JitMemMapSkimData(skim_cache_path, skim_info)
        else:
            # WARNING: memmap gobbles ram up to skim size - see note above
            skim_data = self._open_existing_readonly_memmap_skim_cache(skim_info)
            skim_data = SkimData(skim_data)

        logger.info(
            f"get_skim_data {skim_tag} {type(skim_data).__name__} shape {skim_data.shape}"
        )

        return skim_data



# ActivitySim
# See full license in LICENSE.txt.

import os

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from activitysim.core import inject

from ..vectorize_tour_scheduling import (
    get_previous_tour_by_tourid,
    vectorize_tour_scheduling,
)


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def setup_function():
    output_dir = os.path.join(os.path.dirname(__file__), "output")
    inject.add_injectable("output_dir", output_dir)


def test_vts():

    inject.add_injectable("settings", {})

    # note: need 0 duration tour on one end of day to guarantee at least one available tour
    alts = pd.DataFrame({"start": [1, 1, 2, 3], "end": [1, 4, 5, 6]})
    alts["duration"] = alts.end - alts.start
    inject.add_injectable("tdd_alts", alts)

    current_tour_person_ids = pd.Series(["b", "c"], index=["d", "e"])

    previous_tour_by_personid = pd.Series([2, 2, 1], index=["a", "b", "c"])

    prev_tour_attrs = get_previous_tour_by_tourid(
        current_tour_person_ids, previous_tour_by_personid, alts
    )

    pdt.assert_series_equal(
        prev_tour_attrs.start_previous,
        pd.Series([2, 1], index=["d", "e"], name="start_previous"),
    )

    pdt.assert_series_equal(
        prev_tour_attrs.end_previous,
        pd.Series([5, 4], index=["d", "e"], name="end_previous"),
    )

    tours = pd.DataFrame(
        {
            "person_id": [1, 1, 2, 3, 3],
            "tour_num": [1, 2, 1, 1, 2],
            "tour_type": ["x", "x", "x", "x", "x"],
        }
    )

    persons = pd.DataFrame({"income": [20, 30, 25]}, index=[1, 2, 3])

    inject.add_table("persons", persons)

    spec = pd.DataFrame({"Coefficient": [1.2]}, index=["income"])
    spec.index.name = "Expression"

    inject.add_injectable("check_for_variability", True)

    timetable = inject.get_injectable("timetable")

    tdd_choices = vectorize_tour_scheduling(
        tours,
        persons,
        alts,
        timetable,
        tour_segments={"spec": spec},
        tour_segment_col=None,
        model_settings={},
        chunk_size=0,
        trace_label="test_vts",
    )

    # FIXME - dead reckoning regression
    # there's no real logic here - this is just what came out of the monte carlo
    # note that the result comes out ordered by the nth trips and not ordered
    # by the trip index.  shrug?
    expected = [2, 2, 2, 0, 0]
    assert (tdd_choices.values == expected).all()



# ActivitySim
# See full license in LICENSE.txt.

import hashlib
import logging
from builtins import object, range

import numpy as np
import pandas as pd

from activitysim.core.util import reindex

from .tracing import print_elapsed_time

logger = logging.getLogger(__name__)

# one more than 0xFFFFFFFF so we can wrap using: int64 % _MAX_SEED
_MAX_SEED = 1 << 32
_SEED_MASK = 0xFFFFFFFF


def hash32(s):
    """

    Parameters
    ----------
    s: str

    Returns
    -------
        32 bit unsigned hash
    """
    s = s.encode("utf8")
    h = hashlib.md5(s).hexdigest()
    return int(h, base=16) & _SEED_MASK


class SimpleChannel(object):
    """

    We need to ensure that we generate the same random streams (when re-run or even across
    different simulations.) We do this by generating a random seed for each domain_df row
    that is based on the domain_df index (which implies that generated tables like tours
    and trips are also created with stable, predictable, repeatable row indexes.

    Because we need to generate a distinct stream for each step, we can't just use the
    domain_df index - we need a strategy for handling multiple steps without generating
    collisions between streams (i.e. choosing the same seed for more than one stream.)

    The easiest way to do this would be to use an array of integers to seed the generator,
    with a global seed, a channel seed, a row seed, and a step seed. Unfortunately, seeding
    numpy RandomState with arrays is a LOT slower than with a single integer seed, and
    speed matters because we reseed on-the-fly for every call because creating a different
    RandomState object for each row uses too much memory (5K per RandomState object)

    numpy random seeds are unsigned int32 so there are 4,294,967,295 available seeds.
    That is probably just about enough to distribute evenly, for most cities, depending on the
    number of households, persons, tours, trips, and steps.

    So we use (global_seed + channel_seed + step_seed + row_index) % (1 << 32)
    to get an int32 seed rather than a tuple.

    We do read in the whole households and persons tables at start time, so we could note the
    max index values. But we might then want a way to ensure stability between the test, example,
    and full datasets. I am punting on this for now.
    """

    def __init__(self, channel_name, base_seed, domain_df, step_name):

        self.base_seed = base_seed

        # ensure that every channel is different, even for the same df index values and max_steps
        self.channel_name = channel_name
        self.channel_seed = hash32(self.channel_name)

        self.step_name = None
        self.step_seed = None
        self.row_states = None

        # create dataframe to hold state for every df row
        self.extend_domain(domain_df)
        assert self.row_states.shape[0] == domain_df.shape[0]

        if step_name:
            self.begin_step(step_name)

    def init_row_states_for_step(self, row_states):
        """
        initialize row states (in place) for new step

        with stable, predictable, repeatable row_seeds for that domain_df index value

        See notes on the seed generation strategy in class comment above.

        Parameters
        ----------
        row_states
        """

        assert self.step_name

        if self.step_name and not row_states.empty:

            row_states["row_seed"] = (
                self.base_seed + self.channel_seed + self.step_seed + row_states.index
            ) % _MAX_SEED

            # number of rands pulled this step
            row_states["offset"] = 0

        return row_states

    def extend_domain(self, domain_df):
        """
        Extend or create row_state df by adding seed info for each row in domain_df

        If extending, the index values of new tables must be disjoint so
        there will be no ambiguity/collisions between rows

        Parameters
        ----------
        domain_df : pandas.DataFrame
            domain dataframe with index values for which random streams are to be generated
            and well-known index name corresponding to the channel
        """

        if domain_df.empty:
            logger.warning(
                "extend_domain for channel %s for empty domain_df" % self.channel_name
            )

        # dataframe to hold state for every df row
        row_states = pd.DataFrame(columns=["row_seed", "offset"], index=domain_df.index)

        if self.step_name and not row_states.empty:
            self.init_row_states_for_step(row_states)

        if self.row_states is None:
            self.row_states = row_states
        else:
            # row_states already exists, so we are extending
            # if extending, these should be new rows, no intersection with existing row_states
            assert len(self.row_states.index.intersection(domain_df.index)) == 0
            self.row_states = pd.concat([self.row_states, row_states])

    def begin_step(self, step_name):
        """
        Reset channel state for a new state

        Parameters
        ----------
        step_name : str
            pipeline step name for this step
        """

        assert self.step_name is None

        self.step_name = step_name
        self.step_seed = hash32(self.step_name)

        self.init_row_states_for_step(self.row_states)

        # standard constant to use for choice_for_df instead of fast-forwarding rand stream
        self.multi_choice_offset = None

    def end_step(self, step_name):

        assert self.step_name == step_name

        self.step_name = None
        self.step_seed = None
        self.row_states["offset"] = 0
        self.row_states["row_seed"] = 0

    def _generators_for_df(self, df):
        """
        Python generator function for iterating over numpy prngs (nomenclature collision!)
        seeded and fast-forwarded on-the-fly to the appropriate position in the channel's
        random number stream for each row in df.

        WARNING:
            since we are reusing a single underlying randomstate,
            prng must be called when yielded as generated sequence,
            not serialized and called later after iterator finishes

        Parameters
        ----------
        df : pandas.DataFrame
            dataframe with index values for which random streams are to be generated
            and well-known index name corresponding to the channel
        """

        # assert no dupes
        assert len(df.index.unique()) == len(df.index)

        df_row_states = self.row_states.loc[df.index]

        prng = np.random.RandomState()
        for row in df_row_states.itertuples():

            prng.seed(row.row_seed)

            if row.offset:
                # consume rands
                prng.rand(row.offset)

            yield prng

    def random_for_df(self, df, step_name, n=1):
        """
        Return n floating point random numbers in range [0, 1) for each row in df
        using the appropriate random channel for each row.

        Subsequent calls (in the same step) will return the next rand for each df row

        The resulting array will be the same length (and order) as df
        This method is designed to support alternative selection from a probability array

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Parameters
        ----------
        df : pandas.DataFrame
            df with index name and values corresponding to a registered channel

        n : int
            number of rands desired per df row

        Returns
        -------
        rands : 2-D ndarray
            array the same length as df, with n floats in range [0, 1) for each df row
        """

        assert self.step_name
        assert self.step_name == step_name

        # - reminder: prng must be called when yielded as generated sequence, not serialized
        generators = self._generators_for_df(df)

        rands = np.asanyarray([prng.rand(n) for prng in generators])
        # update offset for rows we handled
        self.row_states.loc[df.index, "offset"] += n
        return rands

    def normal_for_df(self, df, step_name, mu, sigma, lognormal=False):
        """
        Return a floating point random number in normal (or lognormal) distribution
        for each row in df using the appropriate random channel for each row.

        Subsequent calls (in the same step) will return the next rand for each df row

        The resulting array will be the same length (and order) as df
        This method is designed to support alternative selection from a probability array

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Parameters
        ----------
        df : pandas.DataFrame or Series
            df or series with index name and values corresponding to a registered channel

        mu : float or pd.Series or array of floats with one value per df row
        sigma : float or array of floats with one value per df row

        Returns
        -------
        rands : 2-D ndarray
            array the same length as df, with n floats in range [0, 1) for each df row
        """

        assert self.step_name
        assert self.step_name == step_name

        def to_series(x):
            if np.isscalar(x):
                return [x] * len(df)
            elif isinstance(x, pd.Series):
                return x.values
            return x

        # - reminder: prng must be called when yielded as generated sequence, not serialized
        generators = self._generators_for_df(df)

        mu = to_series(mu)
        sigma = to_series(sigma)

        if lognormal:
            rands = np.asanyarray(
                [
                    prng.lognormal(mean=mu[i], sigma=sigma[i])
                    for i, prng in enumerate(generators)
                ]
            )
        else:
            rands = np.asanyarray(
                [
                    prng.normal(loc=mu[i], scale=sigma[i])
                    for i, prng in enumerate(generators)
                ]
            )

        # update offset for rows we handled
        self.row_states.loc[df.index, "offset"] += 1

        return rands

    def choice_for_df(self, df, step_name, a, size, replace):
        """
        Apply numpy.random.choice once for each row in df
        using the appropriate random channel for each row.

        Concatenate the the choice arrays for every row into a single 1-D ndarray
        The resulting array will be of length: size * len(df.index)
        This method is designed to support creation of a interaction_dataset

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        Parameters
        ----------
        df : pandas.DataFrame
            df with index name and values corresponding to a registered channel

        step_name : str
            current step name so we can update row_states seed info

        The remaining parameters are passed through as arguments to numpy.random.choice

        a : 1-D array-like or int
            If an ndarray, a random sample is generated from its elements.
            If an int, the random sample is generated as if a was np.arange(n)
        size : int or tuple of ints
            Output shape
        replace : boolean
            Whether the sample is with or without replacement

        Returns
        -------
        choices : 1-D ndarray of length: size * len(df.index)
            The generated random samples for each row concatenated into a single (flat) array
        """

        assert self.step_name
        assert self.step_name == step_name

        # initialize the generator iterator
        generators = self._generators_for_df(df)

        sample = np.concatenate(
            tuple(prng.choice(a, size, replace) for prng in generators)
        )

        if not self.multi_choice_offset:
            # FIXME - if replace, should we estimate rands_consumed?
            if replace:
                logger.warning("choice_for_df MULTI_CHOICE_FF with replace")
            # update offset for rows we handled
            self.row_states.loc[df.index, "offset"] += size

        return sample


class Random(object):
    def __init__(self):

        self.channels = {}

        # dict mapping df index name to channel name
        self.index_to_channel = {}

        self.step_name = None
        self.step_seed = None
        self.base_seed = 0
        self.global_rng = np.random.RandomState()

    def get_channel_for_df(self, df):
        """
        Return the channel for this df. Channel should already have been loaded/added.

        Parameters
        ----------
        df : pandas.dataframe
            either a domain_df for a channel being added or extended
            or a df for which random values are to be generated
        """

        channel_name = self.index_to_channel.get(df.index.name, None)
        if channel_name is None:
            raise RuntimeError("No channel with index name '%s'" % df.index.name)
        return self.channels[channel_name]

    # step handling

    def begin_step(self, step_name):
        """
        Register that the pipeline has entered a new step and that global and channel streams
        should transition to the new stream.

        Parameters
        ----------
        step_name : str
            pipeline step name
        """

        assert self.step_name is None
        assert step_name is not None

        self.step_name = step_name

        self.step_seed = hash32(step_name)

        seed = [self.base_seed, self.step_seed]
        self.global_rng = np.random.RandomState(seed)

        for c in self.channels:
            self.channels[c].begin_step(self.step_name)

    def end_step(self, step_name):
        """
        This is mostly just for internal consistency checking -
        I'm not sure it serves any useful purpose except to catch "mis-steps" in the pipeline code

        Parameters
        ----------
        step_name : str
            name of current step (just a consistency check)
        """
        assert self.step_name is not None
        assert self.step_name == step_name

        for c in self.channels:
            self.channels[c].end_step(self.step_name)

        self.step_name = None
        self.step_seed = None
        self.global_rng = None

    # channel management

    def add_channel(self, channel_name, domain_df):
        """
        Create or extend a channel for generating random number streams for domain_df.

        We need to be prepared to extend an existing channel because mandatory and non-mandatory
        tours are generated separately by different sub-models, but end up members of a common
        tours channel.

        Parameters
        ----------
        domain_df : pandas.DataFrame
            domain dataframe with index values for which random streams are to be generated
            and well-known index name corresponding to the channel

        channel_name : str
            expected channel name provided as a consistency check

        """

        if channel_name in self.channels:

            assert channel_name == self.index_to_channel[domain_df.index.name]
            logger.debug(
                "Random: extending channel '%s' %s ids"
                % (channel_name, len(domain_df.index))
            )
            channel = self.channels[channel_name]

            channel.extend_domain(domain_df)

        else:
            logger.debug(
                "Adding channel '%s' %s ids" % (channel_name, len(domain_df.index))
            )

            channel = SimpleChannel(
                channel_name, self.base_seed, domain_df, self.step_name
            )

            self.channels[channel_name] = channel
            self.index_to_channel[domain_df.index.name] = channel_name

    def drop_channel(self, channel_name):
        """
        Drop channel that won't be used again (saves memory)

        Parameters
        ----------
        channel_name
        """

        if channel_name in self.channels:
            logger.debug("Dropping channel '%s'" % (channel_name,))
            del self.channels[channel_name]
        else:
            logger.error(
                "drop_channel called with unknown channel '%s'" % (channel_name,)
            )

    def set_base_seed(self, seed=None):
        """
        Like seed for numpy.random.RandomState, but generalized for use with all random streams.

        Provide a base seed that will be added to the seeds of all random streams.
        The default base seed value is 0, so set_base_seed(0) is a NOP

        set_base_seed(1) will (e.g.) provide a different set of random streams than the default
        but will provide repeatable results re-running or resuming the simulation

        set_base_seed(None) will set the base seed to a random and unpredictable integer and so
        provides "fully pseudo random" non-repeatable streams with different results every time

        Must be called before first step (before any channels are added or rands are consumed)

        Parameters
        ----------
        seed : int or None
        """

        if self.step_name is not None or self.channels:
            raise RuntimeError("Can only call set_base_seed before the first step.")

        assert len(list(self.channels.keys())) == 0

        if seed is None:
            self.base_seed = np.random.RandomState().randint(_MAX_SEED)
            logger.debug("Set random seed randomly to %s" % self.base_seed)
        else:
            logger.debug("Set random seed base to %s" % seed)
            self.base_seed = seed

    def get_global_rng(self):
        """
        Return a numpy random number generator for use within current step.

        This method is designed to provide random numbers for uses that do not correspond to
        known channel domains. e.g. to select a subset of households to use for the simulation.

        global_rng is reseeded to a predictable value at the beginning of every step so that
        it behaves repeatably when simulation is resumed or re-run.

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Returns
        -------
        global_rng : numpy.random.RandomState()
            numpy random number generator for use within current step

        """
        assert self.step_name is not None
        return self.global_rng

    def get_external_rng(self, one_off_step_name):
        """
        Return a numpy random number generator for step-independent one_off use

        exists to allow sampling of input tables consistent no matter what step they are called in
        """

        seed = [self.base_seed, hash32(one_off_step_name)]
        return np.random.RandomState(seed)

    def random_for_df(self, df, n=1):
        """
        Return a single floating point random number in range [0, 1) for each row in df
        using the appropriate random channel for each row.

        Subsequent calls (in the same step) will return the next rand for each df row

        The resulting array will be the same length (and order) as df
        This method is designed to support alternative selection from a probability array

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        We assume that we can identify the channel to used based on the name of df.index
        This channel should have already been registered by a call to add_channel (q.v.)

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Parameters
        ----------
        df : pandas.DataFrame
            df with index name and values corresponding to a registered channel

        n : int
            number of rands desired (default 1)

        Returns
        -------
        choices : 1-D ndarray the same length as df
            a single float in range [0, 1) for each row in df
        """

        # FIXME - for tests
        if not self.channels:
            rng = np.random.RandomState(0)
            rands = np.asanyarray([rng.rand(n) for _ in range(len(df))])
            return rands

        channel = self.get_channel_for_df(df)
        rands = channel.random_for_df(df, self.step_name, n)
        return rands

    def normal_for_df(self, df, mu=0, sigma=1, broadcast=False):
        """
        Return a single floating point normal random number in range (-inf, inf) for each row in df
        using the appropriate random channel for each row.

        Subsequent calls (in the same step) will return the next rand for each df row

        The resulting array will be the same length (and order) as df
        This method is designed to support alternative selection from a probability array

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        We assume that we can identify the channel to used based on the name of df.index
        This channel should have already been registered by a call to add_channel (q.v.)

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Parameters
        ----------
        df : pandas.DataFrame
            df with index name and values corresponding to a registered channel

        mu : float or array of floats with one value per df row
        sigma : float or array of floats with one value per df row

        Returns
        -------
        rands : 1-D ndarray the same length as df (or Series with same index as df)
            a single float in lognormal distribution for each row in df
        """

        channel = self.get_channel_for_df(df)

        if broadcast:
            alts_df = df
            df = df.index.unique().to_series()
            rands = channel.normal_for_df(
                df, self.step_name, mu=0, sigma=1, lognormal=False
            )
            rands = reindex(pd.Series(rands, index=df.index), alts_df.index)
            rands = rands * sigma + mu
        else:
            rands = channel.normal_for_df(
                df, self.step_name, mu, sigma, lognormal=False
            )

        return rands

    def lognormal_for_df(self, df, mu, sigma, broadcast=False, scale=False):
        """
        Return a single floating point lognormal random number in range [0, inf) for each row in df
        using the appropriate random channel for each row.

        Note that by default (scale=False) the mean and standard deviation are not the values for
        the distribution itself, but of the underlying normal distribution it is derived from.
        This is perhaps counter-intuitive, but it is the way the numpy standard works,
        and so we are conforming to it here.

        If scale=True, then mu and sigma are the desired mean and standard deviation of the
        lognormal distribution instead of the numpy standard where mu and sigma which are the
        values for the distribution itself, rather than of the underlying normal distribution
        it is derived from.

        Subsequent calls (in the same step) will return the next rand for each df row

        The resulting array will be the same length (and order) as df
        This method is designed to support alternative selection from a probability array

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        We assume that we can identify the channel to used based on the name of df.index
        This channel should have already been registered by a call to add_channel (q.v.)

        If "true pseudo random" behavior is desired (i.e. NOT repeatable) the set_base_seed
        method (q.v.) may be used to globally reseed all random streams.

        Parameters
        ----------
        df : pandas.DataFrame, Series, or Index
            df with index name and values corresponding to a registered channel

        mu : float or array of floats with one value per df row
        sigma : float or array of floats with one value per df row

        Returns
        -------
        rands : 1-D ndarray the same length as df (or Series with same index as df)
            a single float in lognormal distribution for each row in df
        """

        if scale:
            # location = ln(mean/sqrt(1 + std_dev^2/mean^2))
            # scale = sqrt(ln(1 + std_dev^2/mean^2))
            x = 1 + ((sigma * sigma) / (mu * mu))
            mu = np.log(mu / (np.sqrt(x)))
            sigma = np.sqrt(np.log(x))

        if broadcast:
            rands = self.normal_for_df(df, mu=mu, sigma=sigma, broadcast=True)
            rands = np.exp(rands)
        else:
            channel = self.get_channel_for_df(df)
            rands = channel.normal_for_df(
                df, self.step_name, mu=mu, sigma=sigma, lognormal=True
            )

        return rands

    def choice_for_df(self, df, a, size, replace):
        """
        Apply numpy.random.choice once for each row in df
        using the appropriate random channel for each row.

        Concatenate the the choice arrays for every row into a single 1-D ndarray
        The resulting array will be of length: size * len(df.index)
        This method is designed to support creation of a interaction_dataset

        The columns in df are ignored; the index name and values are used to determine
        which random number sequence to to use.

        We assume that we can identify the channel to used based on the name of df.index
        This channel should have already been registered by a call to add_channel (q.v.)

        Parameters
        ----------
        df : pandas.DataFrame
            df with index name and values corresponding to a registered channel

        The remaining parameters are passed through as arguments to numpy.random.choice

        a : 1-D array-like or int
            If an ndarray, a random sample is generated from its elements.
            If an int, the random sample is generated as if a was np.arange(n)
        size : int or tuple of ints
            Output shape
        replace : boolean
            Whether the sample is with or without replacement

        Returns
        -------
        choices : 1-D ndarray of length: size * len(df.index)
            The generated random samples for each row concatenated into a single (flat) array
        """

        # FIXME - for tests
        if not self.channels:
            rng = np.random.RandomState(0)
            choices = np.concatenate(
                tuple(rng.choice(a, size, replace) for _ in range(len(df)))
            )
            return choices

        t0 = print_elapsed_time()
        channel = self.get_channel_for_df(df)
        choices = channel.choice_for_df(df, self.step_name, a, size, replace)
        t0 = print_elapsed_time(
            "choice_for_df for %s rows" % len(df.index), t0, debug=True
        )
        return choices



import base64
import contextlib
import io
import json
import os
import tempfile
import traceback
import zlib

from asv.commands import Command
from asv.console import log
from asv.plugins.snakeviz import SnakevizGui


def benchmark_snakeviz(json_record, benchmark=None):
    """
    A utility to directly display saved profiling data in Snakeviz.

    Parameters
    ----------
    json_record : Path-like
        The archived json file that contains profile data for benchmarks.
    benchmark : str, optional
        The name of the benchmark to display.
    """
    from asv import util

    with open(json_record, "rt") as f:
        json_content = json.load(f)
    profiles = json_content.get("profiles", {})
    if benchmark is None or benchmark not in profiles:
        if profiles:
            log.info("\n\nAvailable profiles:")
            for k in profiles.keys():
                log.info(f"- {k}")
        else:
            log.info(f"\n\nNo profiles stored in {json_record}")
        if benchmark is None:
            return
        raise KeyError()
    profile_data = zlib.decompress(base64.b64decode(profiles[benchmark].encode()))
    prefix = benchmark.replace(".", "__") + "."
    with temp_profile(profile_data, prefix) as profile_path:
        log.info(f"Profiling data cached to {profile_path}")
        import pstats

        prof = pstats.Stats(profile_path)
        prof.strip_dirs().dump_stats(profile_path + "b")
        try:
            SnakevizGui.open_profiler_gui(profile_path + "b")
        except KeyboardInterrupt:
            pass
        except Exception:

            traceback.print_exc()
            input(input("Press Enter to continue..."))
        finally:
            os.remove(profile_path + "b")


@contextlib.contextmanager
def temp_profile(profile_data, prefix=None):
    profile_fd, profile_path = tempfile.mkstemp(prefix=prefix)
    try:
        with io.open(profile_fd, "wb", closefd=True) as fd:
            fd.write(profile_data)

        yield profile_path
    finally:
        os.remove(profile_path)


class ProfileInspector(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "snakeviz",
            help="""Run snakeviz on a particular benchmark that has been profiled.""",
            description="Inspect a benchmark profile",
        )

        parser.add_argument(
            "json_record",
            help="""The json file in the benchmark results to read profile data from.""",
        )

        parser.add_argument(
            "benchmark",
            help="""The benchmark to profile.  Must be a
            fully-specified benchmark name. For parameterized benchmark, it
            must include the parameter combination to use, e.g.:
            benchmark_name(param0, param1, ...)""",
            default=None,
            nargs="?",
        )

        parser.set_defaults(func=cls.run_from_args)

        return parser

    @classmethod
    def run_from_conf_args(cls, conf, args, **kwargs):
        return cls.run(json_record=args.json_record, benchmark=args.benchmark)

    @classmethod
    def run(cls, json_record, benchmark):
        benchmark_snakeviz(json_record, benchmark)



import os
from pathlib import Path
from typing import Collection

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model, P, X
from larch.util import Dict

from .general import (
    apply_coefficients,
    construct_nesting_tree,
    cv_to_ca,
    explicit_value_parameters,
    linear_utility_from_spec,
    remove_apostrophes,
    str_repr,
)


def size_coefficients_from_spec(size_spec):
    size_coef = size_spec.stack().reset_index()
    size_coef.index = size_coef.iloc[:, 0] + "_" + size_coef.iloc[:, 1]
    size_coef = size_coef.loc[size_coef.iloc[:, 2] > 0]
    size_coef["constrain"] = "F"
    one_each = size_coef.groupby("segment").first().reset_index()
    size_coef.loc[one_each.iloc[:, 0] + "_" + one_each.iloc[:, 1], "constrain"] = "T"
    size_coef = size_coef.iloc[:, 2:]
    size_coef.columns = ["value", "constrain"]
    size_coef.index.name = "coefficient_name"
    size_coef["value"] = np.log(size_coef["value"])
    return size_coef


def location_choice_model(
    name="workplace_location",
    edb_directory="output/estimation_data_bundle/{name}/",
    coefficients_file="{name}_coefficients.csv",
    spec_file="{name}_SPEC.csv",
    size_spec_file="{name}_size_terms.csv",
    alt_values_file="{name}_alternatives_combined.csv",
    chooser_file="{name}_choosers_combined.csv",
    settings_file="{name}_model_settings.yaml",
    landuse_file="{name}_landuse.csv",
    return_data=False,
):
    model_selector = name.replace("_location", "")
    model_selector = model_selector.replace("_destination", "")
    model_selector = model_selector.replace("_subtour", "")
    model_selector = model_selector.replace("_tour", "")
    if model_selector == "joint":
        model_selector = "non_mandatory"
    edb_directory = edb_directory.format(name=name)

    def _read_csv(filename, **kwargs):
        filename = filename.format(name=name)
        return pd.read_csv(os.path.join(edb_directory, filename), **kwargs)

    coefficients = _read_csv(
        coefficients_file,
        index_col="coefficient_name",
    )
    spec = _read_csv(spec_file, comment="#")
    alt_values = _read_csv(alt_values_file)
    chooser_data = _read_csv(chooser_file)
    landuse = _read_csv(landuse_file, index_col="zone_id")
    master_size_spec = _read_csv(size_spec_file)

    # remove temp rows from spec, ASim uses them to calculate the other values written
    # to the EDB, but they are not actually part of the utility function themselves.
    spec = spec.loc[~spec.Expression.isna()]
    spec = spec.loc[~spec.Expression.str.startswith("_")].copy()

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file), "r") as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    include_settings = settings.get("include_settings")
    if include_settings:
        include_settings = os.path.join(edb_directory, include_settings)
    if include_settings and os.path.exists(include_settings):
        with open(include_settings, "r") as yf:
            more_settings = yaml.load(
                yf,
                Loader=yaml.SafeLoader,
            )
        settings.update(more_settings)

    CHOOSER_SEGMENT_COLUMN_NAME = settings.get("CHOOSER_SEGMENT_COLUMN_NAME")
    SEGMENT_IDS = settings.get("SEGMENT_IDS")
    if SEGMENT_IDS is None:
        SEGMENTS = settings.get("SEGMENTS")
        if SEGMENTS is not None:
            SEGMENT_IDS = {i: i for i in SEGMENTS}

    SIZE_TERM_SELECTOR = settings.get("SIZE_TERM_SELECTOR", model_selector)

    # filter size spec for this location choice only
    size_spec = (
        master_size_spec.query(f"model_selector == '{SIZE_TERM_SELECTOR}'")
        .drop(columns="model_selector")
        .set_index("segment")
    )
    size_spec = size_spec.loc[:, size_spec.max() > 0]

    size_coef = size_coefficients_from_spec(size_spec)

    indexes_to_drop = [
        "util_size_variable",  # pre-computed size (will be re-estimated)
        "util_size_variable_atwork",  # pre-computed size (will be re-estimated)
        "util_utility_adjustment",  # shadow pricing (ignored in estimation)
        "@df['size_term'].apply(np.log1p)",  # pre-computed size (will be re-estimated)
    ]
    if "Label" in spec.columns:
        indexes_to_drop = [i for i in indexes_to_drop if i in spec.Label.to_numpy()]
        label_column_name = "Label"
    elif "Expression" in spec.columns:
        indexes_to_drop = [
            i for i in indexes_to_drop if i in spec.Expression.to_numpy()
        ]
        label_column_name = "Expression"
    else:
        raise ValueError("cannot find Label or Expression in spec file")

    expression_labels = None
    if label_column_name == "Expression":
        expression_labels = {
            expr: f"variable_label{n:04d}"
            for n, expr in enumerate(spec.Expression.to_numpy())
        }

    # Remove shadow pricing and pre-existing size expression for re-estimation
    spec = spec.set_index(label_column_name).drop(index=indexes_to_drop).reset_index()

    if label_column_name == "Expression":
        spec.insert(0, "Label", spec["Expression"].map(expression_labels))
        alt_values["variable"] = alt_values["variable"].map(expression_labels)
        label_column_name = "Label"

    if name == "trip_destination":
        CHOOSER_SEGMENT_COLUMN_NAME = "primary_purpose"
        primary_purposes = spec.columns[3:]
        SEGMENT_IDS = {pp: pp for pp in primary_purposes}

    chooser_index_name = chooser_data.columns[0]
    x_co = chooser_data.set_index(chooser_index_name)
    x_ca = cv_to_ca(alt_values.set_index([chooser_index_name, alt_values.columns[1]]))

    if CHOOSER_SEGMENT_COLUMN_NAME is not None:
        # label segments with names
        SEGMENT_IDS_REVERSE = {v: k for k, v in SEGMENT_IDS.items()}
        x_co["_segment_label"] = x_co[CHOOSER_SEGMENT_COLUMN_NAME].apply(
            lambda x: SEGMENT_IDS_REVERSE[x]
        )
    else:
        x_co["_segment_label"] = size_spec.index[0]

    # compute total size values by segment
    for segment in size_spec.index:
        total_size_segment = pd.Series(0, index=landuse.index)
        x_co["total_size_" + segment] = 0
        for land_use_field in size_spec.loc[segment].index:
            total_size_segment += (
                landuse[land_use_field] * size_spec.loc[segment, land_use_field]
            )
        x_co["total_size_" + segment] = total_size_segment.loc[
            x_co["override_choice"]
        ].to_numpy()

    # for each chooser, collate the appropriate total size value
    x_co["total_size_segment"] = 0
    for segment in size_spec.index:
        labels = "total_size_" + segment
        rows = x_co["_segment_label"] == segment
        x_co.loc[rows, "total_size_segment"] = x_co[labels][rows]

    # Remove choosers with invalid observed choice (appropriate total size value = 0)
    valid_observed_zone = x_co["total_size_segment"] > 0
    x_co = x_co[valid_observed_zone]
    x_ca = x_ca[x_ca.index.get_level_values(chooser_index_name).isin(x_co.index)]

    # Merge land use characteristics into CA data
    try:
        x_ca_1 = pd.merge(x_ca, landuse, on="zone_id", how="left")
    except KeyError:
        # Missing the zone_id variable?
        # Use the alternative id's instead, which assumes no sampling of alternatives
        x_ca_1 = pd.merge(
            x_ca,
            landuse,
            left_on=x_ca.index.get_level_values(1),
            right_index=True,
            how="left",
        )
    x_ca_1.index = x_ca.index

    # Availability of choice zones
    if "util_no_attractions" in x_ca_1:
        av = (
            x_ca_1["util_no_attractions"]
            .apply(lambda x: False if x == 1 else True)
            .astype(np.int8)
        )
    elif "@df['size_term']==0" in x_ca_1:
        av = (
            x_ca_1["@df['size_term']==0"]
            .apply(lambda x: False if x == 1 else True)
            .astype(np.int8)
        )
    else:
        av = 1

    d = DataFrames(co=x_co, ca=x_ca_1, av=av)

    m = Model(dataservice=d)
    if len(spec.columns) == 4 and all(
        spec.columns == ["Label", "Description", "Expression", "coefficient"]
    ):
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=spec.columns[-1],
            ignore_x=("local_dist",),
        )
    elif (
        len(spec.columns) == 4
        and all(spec.columns[:3] == ["Label", "Description", "Expression"])
        and len(SEGMENT_IDS) == 1
        and spec.columns[3] == list(SEGMENT_IDS.values())[0]
    ):
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col="Label",
            p_col=spec.columns[-1],
            ignore_x=("local_dist",),
        )
    else:
        m.utility_ca = linear_utility_from_spec(
            spec,
            x_col=label_column_name,
            p_col=SEGMENT_IDS,
            ignore_x=("local_dist",),
            segment_id=CHOOSER_SEGMENT_COLUMN_NAME,
        )

    if CHOOSER_SEGMENT_COLUMN_NAME is None:
        assert len(size_spec) == 1
        m.quantity_ca = sum(
            P(f"{i}_{q}") * X(q)
            for i in size_spec.index
            for q in size_spec.columns
            if size_spec.loc[i, q] != 0
        )
    else:
        m.quantity_ca = sum(
            P(f"{i}_{q}")
            * X(q)
            * X(f"{CHOOSER_SEGMENT_COLUMN_NAME}=={str_repr(SEGMENT_IDS[i])}")
            for i in size_spec.index
            for q in size_spec.columns
            if size_spec.loc[i, q] != 0
        )

    apply_coefficients(coefficients, m, minimum=-25, maximum=25)
    apply_coefficients(size_coef, m, minimum=-6, maximum=6)

    m.choice_co_code = "override_choice"

    if return_data:
        return (
            m,
            Dict(
                edb_directory=Path(edb_directory),
                alt_values=alt_values,
                chooser_data=chooser_data,
                coefficients=coefficients,
                landuse=landuse,
                spec=spec,
                size_spec=size_spec,
                master_size_spec=master_size_spec,
                model_selector=model_selector,
                settings=settings,
            ),
        )

    return m


def update_size_spec(model, data, result_dir=Path("."), output_file=None):
    master_size_spec = data.master_size_spec
    size_spec = data.size_spec
    model_selector = data.model_selector

    # Write size coefficients into size_spec
    for c in size_spec.columns:
        for i in size_spec.index:
            param_name = f"{i}_{c}"
            j = (master_size_spec["segment"] == i) & (
                master_size_spec["model_selector"] == model_selector
            )
            try:
                master_size_spec.loc[j, c] = np.exp(model.get_value(param_name))
            except KeyError:
                pass

    # Rescale each row to total 1, not mathematically needed
    # but to maintain a consistent approach from existing ASim
    master_size_spec.iloc[:, 2:] = master_size_spec.iloc[:, 2:].div(
        master_size_spec.iloc[:, 2:].sum(1), axis=0
    )

    if output_file is not None:
        os.makedirs(result_dir, exist_ok=True)
        master_size_spec.reset_index().to_csv(
            result_dir / output_file,
            index=False,
        )

    return master_size_spec


def workplace_location_model(**kwargs):
    unused = kwargs.pop("name", None)
    return location_choice_model(
        name="workplace_location",
        **kwargs,
    )


def school_location_model(**kwargs):
    unused = kwargs.pop("name", None)
    return location_choice_model(
        name="school_location",
        **kwargs,
    )


def atwork_subtour_destination_model(**kwargs):
    unused = kwargs.pop("name", None)
    return location_choice_model(
        name="atwork_subtour_destination",
        **kwargs,
    )


def joint_tour_destination_model(**kwargs):
    # goes with non_mandatory_tour_destination
    unused = kwargs.pop("name", None)
    if "coefficients_file" not in kwargs:
        kwargs["coefficients_file"] = "non_mandatory_tour_destination_coefficients.csv"
    return location_choice_model(
        name="joint_tour_destination",
        **kwargs,
    )


def non_mandatory_tour_destination_model(**kwargs):
    # goes with joint_tour_destination
    unused = kwargs.pop("name", None)
    return location_choice_model(
        name="non_mandatory_tour_destination",
        **kwargs,
    )


def trip_destination_model(**kwargs):
    unused = kwargs.pop("name", None)
    return location_choice_model(
        name="trip_destination",
        **kwargs,
    )
